from enum import Enum
from typing import Any, Dict, List, Literal, Optional, Union

from pydantic import BaseModel, ConfigDict, Field, StrictInt, StrictStr

from .urls import DatasetUrl

# Note that instance types above 4xlarge are undocumented publicly
WHITELISTED_INSTANCE_TYPES = Literal[
    "m5.large",
    "m5.xlarge",
    "m5.2xlarge",
    "m5.4xlarge",
    "m5.8xlarge",
    "m5.12xlarge",
    "m5.16xlarge",
    "r5.large",
    "r5.xlarge",
    "r5.2xlarge",
    "r5.4xlarge",
    "r5.8xlarge",
    "r5.12xlarge",
    "r5.16xlarge",
]


class GetTableBboxRequest(BaseModel):
    path: str
    bbox_minx: float
    bbox_miny: float
    bbox_maxx: float
    bbox_maxy: float

    n_rows: Optional[int] = None
    clip: bool = True
    columns: Optional[List[str]] = None
    buffer: Optional[float] = None
    model_config = ConfigDict(populate_by_name=True)


class StartJobRequest(BaseModel):
    region: Optional[StrictStr] = None
    instance_type: Optional[WHITELISTED_INSTANCE_TYPES] = None
    disk_size_gb: Optional[StrictInt] = None
    additional_env: Optional[List[StrictStr]] = Field(default_factory=list)
    backend: Optional[StrictStr] = None


class ListJobsRequest(BaseModel):
    skip: Optional[StrictInt] = None
    limit: Optional[StrictInt] = None


class UdfType(str, Enum):
    auto = "auto"
    vector_tile = "vector_tile"
    vector_single = "vector_single"
    raster = "raster"
    raster_single = "raster_single"
    app = "app"


class SaveUdfRequest(BaseModel):
    slug: Optional[str] = None
    udf_body: str
    udf_type: UdfType
    allow_public_read: Optional[bool] = None
    allow_public_list: Optional[bool] = None


class ListUdfsRequest(BaseModel):
    skip: Optional[StrictInt] = None
    limit: Optional[StrictInt] = None


class GetPathRequest(BaseModel):
    path: DatasetUrl


class SignPathRequest(BaseModel):
    path: DatasetUrl


class ListPathRequest(BaseModel):
    path: DatasetUrl


class DeletePathRequest(BaseModel):
    path: DatasetUrl
    max_deletion_depth: Union[Optional[int], Literal["unlimited"]]


class ResolvePathRequest(BaseModel):
    path: DatasetUrl


class UploadRequest(BaseModel):
    path: DatasetUrl


class UploadTempRequest(BaseModel):
    extension: Optional[str] = None


class SignUploadRequest(BaseModel):
    content_length: int
    path: DatasetUrl


class SignUploadTempRequest(BaseModel):
    content_length: int
    extension: Optional[str] = None


class ListUdfAccessTokensRequest(BaseModel):
    skip: Optional[StrictInt] = None
    limit: Optional[StrictInt] = None


class CreateUdfAccessTokenRequest(BaseModel):
    udf_email: Optional[str] = None
    udf_slug: Optional[str] = None
    udf_id: Optional[str] = None
    client_id: Optional[str] = None
    cache: bool = True
    metadata_json: Dict[str, Any] = {}
    enabled: bool = True


class UpdateUdfAccessTokenRequest(BaseModel):
    client_id: Optional[str] = None
    cache: Optional[bool] = None
    metadata_json: Optional[Dict[str, Any]] = None
    enabled: Optional[bool] = None


class UpdateSecretRequest(BaseModel):
    value: str



# ruff: noqa: F401


from .api import (
    AnyJobStepConfig,
    Dataset,
    GeospatialPartitionJobStepConfig,
    JobConfig,
    JobStepConfig,
    JoinJobStepConfig,
    JoinSinglefileJobStepConfig,
    MapJobStepConfig,
    NonGeospatialPartitionJobStepConfig,
    PartitionJobStepConfig,
    RootAnyJobStepConfig,
    Table,
    UdfAccessToken,
    UdfAccessTokenList,
    UdfJobStepConfig,
)
from .input import BaseInput, JoinInput, JoinSingleFileInput, MapInput
from .internal import JobResponse, Jobs, RunResponse
from .migrations import migration
from .schema import (
    DataType,
    Field,
    FixedSizeBinaryType,
    FixedSizeListType,
    LargeListType,
    ListType,
    PrimitiveDataType,
    Schema,
    StructType,
)
from .udf import Chunk, ChunkMetadata, GeoPandasUdfV2, Header



from typing import Any, Union

from fused.core._impl._udf_ops_impl import get_step_config_from_server

from .core._impl._reimports import AttrDict


class _PublicUtils:
    """
    A class designed to dynamically access utils of public UDFs, fetching their configuration
    from the server.

    Methods:
        __init__(self, cache_key: Any = None): Initializes the _Public instance with an optional cache key.
        __getattribute__(self, key: str) -> Union[Any, AttrDict]: Attempts to access class attributes, falling back to fetching the attribute's configuration from a server if not found.
        __getitem__(self, key: str) -> AttrDict: Fetches the configuration for the given key from a server and returns a utility object for execution.
    """

    def __init__(self, cache_key: Any = None):
        """
        Initializes the _Public instance with an optional cache key.

        Parameters:
            cache_key (Any, optional): A key used for caching purposes. Defaults to None.
        """
        self._cache_key = cache_key

    def __getattribute__(self, key: str) -> Union[Any, AttrDict]:
        try:
            return super().__getattribute__(key)
        except AttributeError:
            try:
                return self[key]
            # Note that we need to raise an AttributeError, **not a KeyError** so that
            # IPython's _repr_html_ works here
            except KeyError:
                raise AttributeError(
                    f"object of type {type(self).__name__} has no attribute {key}"
                ) from None

    def __getitem__(self, key: str) -> AttrDict:
        step_config = get_step_config_from_server(
            email_or_handle=None,
            slug=key,
            cache_key=self._cache_key,
            _is_public=True,
        )

        return step_config.udf.utils


utils = _PublicUtils()
"""
A module to access utility functions located in the UDF called "common".

They can be imported by other UDFs with `common = fused.public.common`. They contain common operations such as:
- read_shape_zip
- url_to_arr
- get_collection_bbox
- read_tiff_pc
- table_to_tile
- rasterize_geometry


Examples:

    This example shows how to access the `geo_buffer` function from the `common` UDF.
    ```python
    import fused
    import geopandas as gpd

    gdf = gpd.read_file('https://www2.census.gov/geo/tiger/TIGER_RD18/STATE/11_DISTRICT_OF_COLUMBIA/11/tl_rd22_11_bg.zip')
    gdf_buffered = fused.public.common.geo_buffer(gdf, 10)
    ```

    This example shows how to load a table with `table_to_tile`, which efficiently loads a table by filtering and adjusting based on the provided bounding box (bbox) and zoom level.
    ```python
    table_path = "s3://fused-asset/infra/census_bg_us"
    gdf = fused.public.common.table_to_tile(
        bbox, table_path, use_columns=["GEOID", "geometry"], min_zoom=12
    )
    ```

    This example shows how to use `rasterize_geometry` to place an input geometry within an image array.
    ```python
    geom_masks = [
        rasterize_geometry(geom, arr.shape[-2:], transform) for geom in gdf.geometry
    ]
    ```

Public UDFs are listed at https://github.com/fusedio/udfs/tree/main/public
"""



from fused._optional_deps import (
    GPD_GEODATAFRAME,
    HAS_GEOPANDAS,
    HAS_MERCANTILE,
    HAS_SHAPELY,
    MERCANTILE_TILE,
    SHAPELY_GEOMETRY,
)


def estimate_zoom(bounds) -> int:
    """
    Estimate the zoom level for a given bounding box.

    This method returns the zoom level at which a tile exists that, potentially
    shifted slightly, fully covers the bounding box.

    Args:
        bounds: A list of 4 coordinates (minx, miny, maxx, maxy), a
            GeoDataFrame or Shapely geometry, or a mercantile Tile.

    Returns:
        The estimated zoom level (0-20).

    """
    if HAS_GEOPANDAS and isinstance(bounds, GPD_GEODATAFRAME):
        bounds = bounds.total_bounds
    elif HAS_SHAPELY and isinstance(bounds, SHAPELY_GEOMETRY):
        bounds = bounds.bounds
    elif HAS_MERCANTILE and isinstance(bounds, MERCANTILE_TILE):
        return bounds.z
    elif not isinstance(bounds, list):
        raise TypeError(f"Invalid bounds type: {type(bounds)}")

    if not HAS_MERCANTILE:
        raise ImportError("This function requires the mercantile package.")
    import mercantile

    minx, miny, maxx, maxy = bounds

    centroid = (minx + maxx) / 2, (miny + maxy) / 2
    width = (maxx - minx) - 1e-11
    height = (maxy - miny) - 1e-11

    for z in range(20, 0, -1):
        tile = mercantile.tile(*centroid, zoom=z)
        west, south, east, north = mercantile.bounds(tile)
        if width <= (east - west) and height <= (north - south):
            break

    return z



"""
This module is responsible for creating sample UDF contexts for each algorithm (join, map, etc) for development purposes.
"""
from enum import Enum
from io import BytesIO
from typing import Dict, List, Literal, Optional
from zipfile import ZipFile

import geopandas as gpd
import pandas as pd
import pyarrow.parquet as pq
from pydantic import BaseModel

from fused_batch.models import JoinInput, JoinSingleFileInput, MapInput
from fused_batch.models.api import AnyJobStepConfig
from fused_batch.models.udf import Chunk, ChunkMetadata, Chunks


class ArgsType(str, Enum):
    MAP = "map"
    JOIN = "join"
    JOIN_SINGLEFILE = "join_singlefile"


class ArgsMetadata(BaseModel):
    args_version: Literal["1"] = "1"
    args_type: ArgsType
    step_config: AnyJobStepConfig


def extract_sidecar_inputs(zf: ZipFile, prefix: str) -> Dict[str, bytes]:
    prefix = prefix.rstrip("/")
    sidecar_input_buffers: Dict[str, bytes] = {}
    sidecar_input_paths = [
        info for info in zf.filelist if info.filename.startswith(prefix)
    ]
    for sidecar_input in sidecar_input_paths:
        table_name = sidecar_input.filename.split("/")[-1]
        sidecar_input_buffers[table_name] = zf.read(sidecar_input)

    return sidecar_input_buffers


def extract_chunk(zf: ZipFile, prefix: str, *, n_rows: Optional[int] = None) -> Chunk:
    prefix = prefix.rstrip("/")
    with zf.open(f"{prefix}/data") as f:
        metadata = pq.read_metadata(f)
        if b"geo" in metadata.metadata:
            data = gpd.read_parquet(f)
        else:
            data = pd.read_parquet(f)

    data = data[:n_rows]

    metadata = ChunkMetadata.model_validate_json(
        zf.read(f"{prefix}/row_group_metadata")
    )
    sidecar_inputs = extract_sidecar_inputs(zf, f"{prefix}/sidecar_inputs")

    return Chunk(data=data, metadata=metadata, sidecar=sidecar_inputs)


def zip_to_map_args(buffer: bytes, *, n_rows: Optional[int] = None) -> MapInput:
    with BytesIO(buffer) as bio, ZipFile(bio, mode="r") as zf:
        args_metadata = ArgsMetadata.model_validate_json(zf.read("fused_args/metadata"))
        assert args_metadata.args_version == "1"
        assert args_metadata.args_type == ArgsType.MAP

        chunk = extract_chunk(zf, "map", n_rows=n_rows)
        return MapInput(
            step_config=args_metadata.step_config,
            data=chunk.data,
            metadata=chunk.metadata,
            sidecar=chunk.sidecar,
        )


def zip_to_join_args(buffer: bytes, *, n_rows: Optional[int] = None) -> JoinInput:
    with BytesIO(buffer) as bio, ZipFile(bio, mode="r") as zf:
        args_metadata = ArgsMetadata.model_validate_json(zf.read("fused_args/metadata"))
        assert args_metadata.args_version == "1"
        assert args_metadata.args_type == ArgsType.JOIN

        left_chunk = extract_chunk(zf, "join", n_rows=n_rows)

        num_rights = int(zf.read("join/num_rights").decode("utf-8"))
        right_chunks: List[Chunk] = []
        for i in range(num_rights):
            prefix = f"join/right/{i}"
            right_chunk = extract_chunk(zf, prefix)
            right_chunks.append(right_chunk)

        return JoinInput(
            step_config=args_metadata.step_config,
            left=left_chunk,
            right=Chunks.from_list(right_chunks),
        )


def zip_to_join_singlefile_args(
    buffer: bytes, *, n_rows: Optional[int] = None
) -> JoinSingleFileInput:
    with BytesIO(buffer) as bio, ZipFile(bio, mode="r") as zf:
        args_metadata = ArgsMetadata.model_validate_json(zf.read("fused_args/metadata"))
        assert args_metadata.args_version == "1"
        assert args_metadata.args_type == ArgsType.JOIN_SINGLEFILE

        left_chunk = extract_chunk(zf, "join", n_rows=n_rows)

        with zf.open("join/right/data") as f:
            metadata = pq.read_metadata(f)
            if b"geo" in metadata.metadata:
                right_data = gpd.read_parquet(f)
            else:
                right_data = pd.read_parquet(f)

        return JoinSingleFileInput(
            step_config=args_metadata.step_config, left=left_chunk, right=right_data
        )



from typing import TYPE_CHECKING, Any, Dict, List, Optional, Tuple, Union

import geopandas as gpd
import pandas as pd
import pyarrow as pa
import shapely
from pydantic import BaseModel, ConfigDict, RootModel, StrictStr

if TYPE_CHECKING:
    from pydantic.typing import ReprArgs


class ChunkMetadata(BaseModel):
    file_id: str
    """The identifier of the file that contains this chunk."""

    chunk_id: int
    """The identifier of this chunk inside the file."""

    bbox_minx: float
    """The west-most coordinate of this bounding box."""

    bbox_miny: float
    """The southern-most coordinate of this bounding box."""

    bbox_maxx: float
    """The east-most coordinate of this bounding box."""

    bbox_maxy: float
    """The northern-most coordinate of this bounding box."""

    sum_area: float
    """The sum of all area of all geometries in this chunk.

    !!! note

        Area is currently computed in the WGS84 coordinate system, so it should only be used as a heuristic.
    """

    sum_length: float
    """The sum of all lengths of all geometries in this chunk.

    !!! note

        Length is currently computed in the WGS84 coordinate system, so it should only be used as a heuristic.
    """

    # These are set to optional with a None default because older datasets might not
    # have utm area/length defined
    sum_area_utm: Optional[float] = None
    """The sum of all geometries' area in UTM"""

    sum_length_utm: Optional[float] = None
    """The sum of all geometries' length in UTM"""

    num_coords: int
    """The sum of the number of coordinates among all geometries in this chunk.
    """

    num_rows: int
    """The number of rows in this chunk"""

    def to_box(self) -> shapely.Polygon:
        """Returns a Shapely polygon representing the bounding box of this chunk"""
        return shapely.box(
            xmin=self.bbox_minx,
            ymin=self.bbox_miny,
            xmax=self.bbox_maxx,
            ymax=self.bbox_maxy,
        )


class SidecarDict:
    def __init__(self, wrapped_dict: Dict[StrictStr, bytes]):
        self.wrapped_dict = wrapped_dict

    def __repr__(self) -> str:
        sidecar_details = ", ".join(
            [f"{key}: {len(val)}" for key, val in self.wrapped_dict.items()]
        )
        return f"SidecarDict({sidecar_details})"


def _wrap_sidecar_dict(to_wrap: Dict[StrictStr, bytes]) -> SidecarDict:
    if to_wrap is not None:
        return SidecarDict(to_wrap)
    else:
        return to_wrap


ChunkData = Union[pa.Table, pd.DataFrame, gpd.GeoDataFrame]


class Chunk(BaseModel):
    data: ChunkData
    """The data object contained within this chunk.

    This will either be a [`pandas.DataFrame`][pandas.DataFrame],
    [`geopandas.GeoDataFrame`][geopandas.GeoDataFrame], or
    [`pyarrow.Table`][pyarrow.Table], depending on the type of user-defined function in
    use.
    """

    metadata: ChunkMetadata
    """The [ChunkMetadata][fused.models.udf.common.ChunkMetadata] describing this chunk."""

    sidecar: Optional[Dict[StrictStr, bytes]] = None

    def _repr_html_(self) -> str:
        data_repr = (
            f"{self.data._repr_html_()}"
            if hasattr(self.data, "_repr_html_")
            else f"{self.data}"
        )
        return f"""
        {data_repr}<br/>
        Metadata: <code>{self.metadata}</code><br/>
        {f'Has sidecar <code>[{", ".join(self.sidecar.keys())}]</code>' if self.sidecar else ''}
        """

    def __repr_args__(self) -> "ReprArgs":
        args = super().__repr_args__()

        def _transform_chunk_repr(
            arg: Tuple[Optional[str], Any]
        ) -> Tuple[Optional[str], Any]:
            if arg[0] == "sidecar" and arg[1] is not None:
                # Wrap the sidecar dict in order to avoid dumping a bunch of bytes as output
                # TODO: It would be nice if this were the type of the field
                return (arg[0], _wrap_sidecar_dict(arg[1]))
            else:
                return arg

        return [_transform_chunk_repr(arg) for arg in args]

    model_config = ConfigDict(arbitrary_types_allowed=True)


class Chunks(RootModel[List[Chunk]]):
    def _repr_html_(self) -> str:
        text = "<ul>"
        for chunk_idx, chunk in enumerate(self):
            text += f"<li><h3>{chunk_idx}</h3>{chunk._repr_html_()}</li>"
        text += "</ul>"
        return text

    @property
    def data(self) -> ChunkData:
        return self.concat_data()

    def concat_data(self, **kwargs) -> ChunkData:
        if len(self.root) == 0:
            # TODO: If this could be a different type -- pa.Table, for example -- that needs
            # to be passed in to Chunks.from_list.
            return pd.DataFrame()

        first = self.root[0].data
        if isinstance(first, pd.DataFrame):
            assert all(isinstance(chunk.data, pd.DataFrame) for chunk in self.root)
            return pd.concat([chunk.data for chunk in self.root], **kwargs)
        else:
            return pa.concat_tables([chunk.data for chunk in self.root], **kwargs)

    def __iter__(self):
        return iter(self.root)

    def __getitem__(self, item):
        return self.root[item]

    def __len__(self):
        return len(self.root)

    @classmethod
    def from_list(cls, chunks: List[Chunk]) -> "Chunks":
        return cls(root=chunks)



import sys
import warnings
from functools import partial
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Coroutine,
    Dict,
    Literal,
    Optional,
    Union,
    overload,
)

if TYPE_CHECKING:
    import geopandas as gpd
    import pandas as pd
    import xarray as xr

from fused._udf.decorators import _parse_ttl
from fused.models.api.udf_access_token import is_udf_token
from fused.models.udf._eval_result import UdfEvaluationResult
from fused.types import UdfRuntimeError
from fused.warnings import FusedDeprecationWarning, FusedIgnoredWarning, FusedWarning

from .core import (
    run_file,
    run_file_async,
    run_shared_file,
    run_shared_file_async,
    run_shared_tile,
    run_shared_tile_async,
    run_tile,
    run_tile_async,
)
from .core._impl._realtime_ops_impl import default_run_engine
from .core._impl._reimports import GeoPandasUdfV2, UdfAccessToken, UdfJobStepConfig

ResultType = Union["xr.Dataset", "pd.DataFrame", "gpd.GeoDataFrame"]


@overload
def run(
    udf: Union[str, None, UdfJobStepConfig, GeoPandasUdfV2, UdfAccessToken] = None,
    *args,
    x: Optional[int] = None,
    y: Optional[int] = None,
    z: Optional[int] = None,
    sync: bool = False,
    engine: Optional[Literal["remote", "local"]] = None,
    type: Optional[Literal["tile", "file"]] = None,
    max_retry: int = 0,
    cache_max_age: Optional[str] = None,
    parameters: Optional[Dict[str, Any]] = None,
    _include_log: Optional[bool] = False,
    _return_response: Optional[bool] = False,
    **kw_parameters,
) -> Coroutine[ResultType, None, None]:
    ...


@overload
def run(
    udf: Union[str, None, UdfJobStepConfig, GeoPandasUdfV2, UdfAccessToken] = None,
    *args,
    x: Optional[int] = None,
    y: Optional[int] = None,
    z: Optional[int] = None,
    sync: bool = True,
    engine: Optional[Literal["remote", "local"]] = None,
    type: Optional[Literal["tile", "file"]] = None,
    max_retry: int = 0,
    cache_max_age: Optional[str] = None,
    parameters: Optional[Dict[str, Any]] = None,
    _include_log: Optional[bool] = False,
    _return_response: Optional[bool] = True,
    **kw_parameters,
) -> UdfEvaluationResult:
    ...


def run(
    udf: Union[str, None, UdfJobStepConfig, GeoPandasUdfV2, UdfAccessToken] = None,
    *args,
    x: Optional[int] = None,
    y: Optional[int] = None,
    z: Optional[int] = None,
    sync: bool = True,
    engine: Optional[Literal["remote", "local"]] = None,
    type: Optional[Literal["tile", "file"]] = None,
    max_retry: int = 0,
    cache_max_age: Optional[str] = None,
    parameters: Optional[Dict[str, Any]] = None,
    _include_log: Optional[bool] = False,
    _return_response: Optional[bool] = False,
    **kw_parameters,
) -> ResultType:
    """
    Executes a user-defined function (UDF) with various execution and input options.

    This function supports executing UDFs in different environments (local or remote),
    with different types of inputs (tile coordinates, geographical bounding boxes, etc.), and
    allows for both synchronous and asynchronous execution. It dynamically determines the execution
    path based on the provided parameters.

    Args:
        udf (str, GeoPandasUdfV2 or UdfJobStepConfig): the UDF to execute.
            The UDF can be specified in several ways:
            - A string representing a UDF name or UDF shared token.
            - A UDF object.
            - A UdfJobStepConfig object for detailed execution configuration.
        x, y, z: Tile coordinates for tile-based UDF execution.
        sync: If True, execute the UDF synchronously. If False, execute asynchronously.
        engine: The execution engine to use ('remote' or 'local').
        type: The type of UDF execution ('tile' or 'file').
        max_retry: The maximum number of retries to attempt if the UDF fails.
            By default does not retry.
        cache_max_age: The maximum age when returning a result from the cache.
        parameters: Additional parameters to pass to the UDF.
        **kw_parameters: Additional parameters to pass to the UDF.

    Raises:
        ValueError: If the UDF is not specified or is specified in more than one way.
        TypeError: If the first parameter is not of an expected type.
        Warning: Various warnings are issued for ignored parameters based on the execution path chosen.

    Returns:
        The result of the UDF execution, which varies based on the UDF and execution path.

    Examples:


        Run a UDF saved in the Fused system:
        ```py
        fused.run("username@fused.io/my_udf_name")
        ```

        Run a UDF saved in GitHub:
        ```py
        loaded_udf = fused.load("https://github.com/fusedio/udfs/tree/main/public/Building_Tile_Example")
        fused.run(loaded_udf, bbox=bbox)
        ```

        Run a UDF saved in a local directory:
        ```py
        loaded_udf = fused.load("/Users/local/dir/Building_Tile_Example")
        fused.run(loaded_udf, bbox=bbox)
        ```

    Note:
        This function dynamically determines the execution path and parameters based on the inputs.
        It is designed to be flexible and support various UDF execution scenarios.
    """
    from fused._optional_deps import HAS_GEOPANDAS, HAS_MERCANTILE, HAS_SHAPELY

    job_step: Optional[UdfJobStepConfig] = None
    token: Optional[str] = None
    udf_email: Optional[str] = None
    udf_name: Optional[str] = None

    if "token" in kw_parameters:
        token = kw_parameters.pop("token")
        if udf is not None:
            warnings.warn(
                "token parameter is being ignored in favor of the first positional parameter.",
                FusedIgnoredWarning,
            )
        else:
            udf = token
        warnings.warn(
            "The 'token' keyword is deprecated. You can pass the token as the first "
            "argument instead (i.e. replace 'fused.run(token=<token>)' with "
            "'fused.run(<token>)')",
            FusedDeprecationWarning,
        )

    elif "udf_email" in kw_parameters and "udf_name" in kw_parameters:
        udf_email = kw_parameters.pop("udf_email")
        udf_name = kw_parameters.pop("udf_name")
        if udf is not None:
            warnings.warn(
                "udf_email parameter is being ignored in favor of the first positional parameter.",
                FusedIgnoredWarning,
            )
        else:
            udf = f"{udf_email}/{udf_name}"
            warnings.warn(
                "The 'udf_email' and 'udf_name' keywords are deprecated. You can pass "
                "the email and name as the first argument instead (i.e. replace "
                "'fused.run(udf_email=<email>, udf_name=<name>)' with "
                "'fused.run(\"<email>/<name>\")')",
                FusedDeprecationWarning,
            )

    elif args:
        if len(args) > 1:
            raise TypeError(
                "run() takes from 0 to 2 positional arguments but 3 were given"
            )
        udf_name = args[0]
        udf = f"{udf}/{udf_name}"
        warnings.warn(
            "The separate 'udf_email' and 'udf_name' arguments are deprecated. You can "
            "pass the email and name as the first argument instead (i.e. replace "
            "'fused.run(<email>, <name>)' with 'fused.run(\"<email>/<name>\")')",
            FusedDeprecationWarning,
        )

    if udf is None:
        raise ValueError("No UDF specified")

    if isinstance(udf, UdfJobStepConfig):
        job_step = udf
        udf_storage = "local_job_step"
    elif isinstance(udf, GeoPandasUdfV2):
        job_step = UdfJobStepConfig(udf=udf)
        udf_storage = "local_job_step"
    elif isinstance(udf, UdfAccessToken):
        token = udf.token
        udf_storage = "token"
    elif isinstance(udf, str):
        if "/" in udf:
            udf_email, udf_name = udf.split("/", maxsplit=1)
            udf_storage = "saved"
        elif "@" in udf:
            udf_email = udf
            udf_storage = "saved"
        elif is_udf_token(udf):
            token = udf
            udf_storage = "token"
        else:
            # This will actually be the udf name, not the user's email
            udf_email = udf
            udf_storage = "saved"
    else:
        raise TypeError(
            "Could not detect UDF from first parameter. It should be a string, UdfJobStepConfig, or BaseUdf object."
        )

    if engine == "realtime":
        engine = "remote"
    elif engine == "batch":
        warnings.warn(
            "The 'batch' engine option is deprecated. Use fused.submit() instead.",
            FusedDeprecationWarning,
            stacklevel=2,
        )
    elif engine is None:
        if udf_storage in ["token", "saved"]:
            engine = "remote"
        else:
            engine = default_run_engine()
    elif engine not in ("local", "remote"):
        raise ValueError("Invalid engine specified. Must be 'local' or 'remote'.")

    local_tile_bbox: Optional["gpd.GeoDataFrame"] = None
    xyz_ignored = False
    if (
        x is not None
        and y is not None
        and z is not None
        and HAS_MERCANTILE
        and HAS_GEOPANDAS
        and HAS_SHAPELY
    ):
        import geopandas as gpd
        import mercantile
        import shapely

        tile_bounds = mercantile.bounds(x, y, z)
        local_tile_bbox = gpd.GeoDataFrame(
            {"x": [x], "y": [y], "z": [z]},
            geometry=[shapely.box(*tile_bounds)],
            crs=4326,
        )
    elif x is not None or y is not None or z is not None:
        xyz_ignored = True

    if x is not None and y is not None and z is not None:
        if type is None:
            type = "tile"
        elif type != "tile":
            warnings.warn(
                FusedIgnoredWarning(
                    "x, y, z specified but UDF type is not 'tile', so they will be ignored"
                ),
            )
    else:
        if type is None:
            type = "file"
        elif type != "file":
            raise ValueError(
                "x, y, z not specified but type is 'tile', which is an invalid configuration. You must specify x, y, and z."
            )

    parameters = {
        **kw_parameters,
        **(parameters if parameters is not None else {}),
        **{"_include_log": True},
    }
    cache_max_age = _parse_ttl(cache_max_age)

    dispatch: dict[
        tuple[
            Literal["sync", "async"],
            Literal["tile", "file"],
            Literal["saved", "token", "local_job_step"],
            Literal["remote", "local", "batch"],
        ],
        Optional[Callable],
    ] = {
        # Saved UDF
        ("sync", "tile", "saved", "remote"): partial(
            run_tile,
            udf_email,
            udf_name,
            x=x,
            y=y,
            z=z,
            cache_max_age=cache_max_age,
            **parameters,
        ),
        ("async", "tile", "saved", "remote"): partial(
            run_tile_async,
            udf_email,
            udf_name,
            x=x,
            y=y,
            z=z,
            cache_max_age=cache_max_age,
            **parameters,
        ),
        ("sync", "file", "saved", "remote"): partial(
            run_file,
            udf_email,
            udf_name,
            cache_max_age=cache_max_age,
            **parameters,
        ),
        ("async", "file", "saved", "remote"): partial(
            run_file_async,
            udf_email,
            udf_name,
            cache_max_age=cache_max_age,
            **parameters,
        ),
        # shared UDF token
        ("sync", "tile", "token", "remote"): partial(
            run_shared_tile,
            token,
            x=x,
            y=y,
            z=z,
            cache_max_age=cache_max_age,
            **parameters,
        ),
        ("async", "tile", "token", "remote"): partial(
            run_shared_tile_async,
            token,
            x=x,
            y=y,
            z=z,
            cache_max_age=cache_max_age,
            **parameters,
        ),
        ("sync", "file", "token", "remote"): partial(
            run_shared_file, token, cache_max_age=cache_max_age, **parameters
        ),
        ("async", "file", "token", "remote"): partial(
            run_shared_file_async, token, cache_max_age=cache_max_age, **parameters
        ),
        # Local job step, which includes locally held code
        ("sync", "tile", "local_job_step", "remote"): lambda: job_step.run_tile(
            x=x, y=y, z=z, cache_max_age=cache_max_age, **parameters
        ),
        (
            "async",
            "tile",
            "local_job_step",
            "remote",
        ): lambda: job_step.run_tile_async(
            x=x, y=y, z=z, cache_max_age=cache_max_age, **parameters
        ),
        ("sync", "file", "local_job_step", "remote"): lambda: job_step.run_file(
            cache_max_age=cache_max_age, **parameters
        ),
        (
            "async",
            "file",
            "local_job_step",
            "remote",
        ): lambda: job_step.run_file_async(**parameters),
        ("sync", "tile", "local_job_step", "local"): lambda: job_step.run_local(
            local_tile_bbox, **parameters, _return_response=_return_response
        ),
        ("async", "tile", "local_job_step", "local"): lambda: job_step.run_local(
            local_tile_bbox, **parameters
        ),
        ("sync", "file", "local_job_step", "local"): lambda: job_step.run_local(
            **parameters, _return_response=_return_response
        ),
        ("async", "file", "local_job_step", "local"): lambda: job_step.run_local(
            **parameters
        ),
        ("sync", "tile", "local_job_step", "batch"): None,
        ("sync", "file", "local_job_step", "batch"): lambda: job_step.set_udf(
            job_step.udf, parameters=parameters
        ).run_remote(),
    }

    if xyz_ignored:
        if x is None or y is None or z is None:
            warnings.warn(
                FusedIgnoredWarning(
                    "x, y, z arguments will be ignored because one of them is None"
                ),
            )
        elif engine == "local":
            # This warning doesn't matter on realtime because we will just put the x/y/z into the URL
            warnings.warn(
                FusedIgnoredWarning(
                    "x, y, z arguments will be ignored because the following packages were not all found: mercantile shapely geopandas"
                ),
            )

    dispatch_params = ("sync" if sync else "async", type, udf_storage, engine)

    # Ellipsis is the sentinal value for not found in the dictionary at all
    fn = dispatch.get(dispatch_params, ...)

    if fn is Ellipsis:
        if udf_storage == "token" and engine != "remote":
            raise ValueError("UDF tokens can only be called on the 'remote' engine.")
        elif udf_storage == "saved" and engine != "remote":
            raise ValueError(
                "Saved UDFs can only be called on the 'remote' engine. To use another engine, load the UDF locally first."
            )
        else:
            raise ValueError(
                f"Could not determine how to call with settings: {dispatch_params}"
            )
    if fn is None:
        raise ValueError(f"Call type is not yet implemented: {dispatch_params}")
    if not sync and udf_storage not in ["saved", "token"]:
        raise ValueError(
            "Asynchronous execution is only supported for saved UDFs specifed by name or token."
        )

    n_retries = 0
    while n_retries <= max_retry:
        try:
            udf_eval_result = fn()
            break
        except Exception as exc:
            if n_retries >= max_retry:
                raise exc
            n_retries += 1
            warnings.warn(
                f"UDF execution failed, retrying (error: {exc})", FusedWarning
            )
            continue

    # Nested and remote UDF calls will return UdfEvaluationResult.
    # merge the stdout/stderr from fused.run() into running environment,
    # unless _include_log=True, then return the UdfEvaluationResult object.
    if _return_response:
        return udf_eval_result
    if isinstance(udf_eval_result, UdfEvaluationResult):
        if udf_eval_result.stdout:
            sys.stdout.write(udf_eval_result.stdout)
        if udf_eval_result.stderr:
            sys.stderr.write(udf_eval_result.stderr)
        if udf_eval_result.error_message is not None:
            raise UdfRuntimeError(
                udf_eval_result.error_message,
                child_exception_class=udf_eval_result.exception_class,
            )
        if _include_log:
            return udf_eval_result
        else:
            return udf_eval_result.data
    return udf_eval_result



from typing import Optional, Union

from .api.dataset import Dataset, Table
from .internal.dataset import (
    AnyDatasetInput,
    AnyDatasetOutput,
    DatasetInput,
    DatasetInputBase,
    DatasetInputV2,
    DatasetOutputBase,
    DatasetOutputV2,
)
from .urls import DatasetUrl

CoerceableToDatasetInput = Union[str, AnyDatasetInput, Dataset, Table, DatasetUrl]

CoerceableToDatasetOutput = Union[str, AnyDatasetOutput]

CoerceableToDataset = Union[str, DatasetInput, Dataset, DatasetUrl]


def _object_to_dataset_input(
    dataset: CoerceableToDatasetInput, *, allow_none: bool = False
) -> Optional[AnyDatasetInput]:
    if isinstance(dataset, Dataset):
        return DatasetInput(base_path=dataset.base_path)
    elif isinstance(dataset, Table):
        input = DatasetInputV2.from_table_url(url=dataset.url)
        input._project_url = (
            dataset._project_url if hasattr(dataset, "_project_url") else None
        )
        return input
    elif isinstance(dataset, str):
        return DatasetInputV2.from_table_url(url=dataset)
    elif isinstance(dataset, DatasetInputBase):
        return dataset
    else:
        if allow_none:
            return None
        assert False, f"Unknown type for dataset input: {type(dataset)}"


def _object_to_dataset_output(
    dataset: Optional[CoerceableToDatasetOutput],
) -> AnyDatasetOutput:
    if dataset is None:
        return DatasetOutputV2()
    elif isinstance(dataset, str):
        return DatasetOutputV2(url=dataset)
    elif isinstance(dataset, DatasetOutputBase):
        return dataset
    else:
        assert False, f"Unknown type for dataset output: {type(dataset)}"



from pathlib import Path
from typing import Any, Union

from .core import load_udf_from_fused, load_udf_from_github
from .core._impl._reimports import AnyBaseUdf, load_job


def load(url_or_udf: Union[str, Path], /, *, cache_key: Any = None) -> AnyBaseUdf:
    """
    Loads a UDF from various sources including GitHub URLs,
    local files, or directories, and a Fused platform-specific identifier.

    This function supports loading UDFs from a GitHub repository URL, a local file path,
    a directory containing UDF definitions, or a Fused platform-specific identifier
    composed of an email and UDF name. It intelligently determines the source type based
    on the format of the input and retrieves the UDF accordingly.

    Args:
        url_or_udf: A string or Path object representing the location of the UDF. This can be
            a GitHub URL starting with "https://github.com", a local file path, a directory
            containing one or more UDFs, or a Fused platform-specific identifier in the
            format "email/udf_name".
        cache_key: An optional key used for caching the loaded UDF. If provided, the function
            will attempt to load the UDF from cache using this key before attempting to
            load it from the specified source. Defaults to None, indicating no caching.

    Returns:
        AnyBaseUdf: An instance of the loaded UDF.

    Raises:
        FileNotFoundError: If a local file or directory path is provided but does not exist.
        ValueError: If the URL or Fused platform-specific identifier format is incorrect or
            cannot be parsed.
        Exception: For errors related to network issues, file access permissions, or other
            unforeseen errors during the loading process.

    Examples:
        Load a UDF from a GitHub URL:
        ```py
        udf = fused.load("https://github.com/fusedio/udfs/tree/main/public/REM_with_HyRiver/")
        ```

        Load a UDF from a local file:
        ```py
        udf = fused.load("/localpath/REM_with_HyRiver/")
        ```

        Load a UDF using a Fused platform-specific identifier:
        ```py
        udf = fused.load("username@fused.io/REM_with_HyRiver")
        ```
    """
    maybe_path = Path(url_or_udf)
    if url_or_udf.startswith("https://github.com"):
        return load_udf_from_github(url_or_udf, cache_key=cache_key)
    elif maybe_path.exists() and maybe_path.is_file():
        job = load_job(maybe_path)
        return job.steps[0].udf
    elif maybe_path.exists() and maybe_path.is_dir():
        job = load_job(maybe_path)
        return job.steps[0].udf
    else:
        email, udf_name = url_or_udf.split("/", maxsplit=1)
        return load_udf_from_fused(email, udf_name, cache_key=cache_key)



from __future__ import annotations

from contextlib import contextmanager
from contextvars import ContextVar
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from fused_batch.models.udf import AnyBaseUdf


noop_decorators: ContextVar[bool] = ContextVar("noop_decorator", default=False)
decorator_src_override: ContextVar[str | None] = ContextVar(
    "decorator_src_override", default=None
)
decorator_udf_override: ContextVar[AnyBaseUdf | None] = ContextVar(
    "decorator_udf_override", default=None
)


@contextmanager
def noop_decorators_context(val: bool):
    token = noop_decorators.set(val)
    try:
        yield token
    finally:
        noop_decorators.reset(token)


# Provides the UDF's code to the fused.udf decorator via the global context (above).
#
# Injecting the UDF code back inside the run function.
@contextmanager
def decorator_src_override_context(val: str):
    token = decorator_src_override.set(val)
    try:
        yield token
    finally:
        decorator_src_override.reset(token)


@contextmanager
def decorator_udf_override_context(udf: AnyBaseUdf):
    token = decorator_udf_override.set(udf)
    try:
        yield token
    finally:
        decorator_udf_override.reset(token)



from __future__ import annotations

import json
import os
import tempfile
import uuid
import warnings
import zipfile
from contextlib import contextmanager
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    Any,
    BinaryIO,
    Dict,
    Iterable,
    List,
    Literal,
    Optional,
    Sequence,
    Tuple,
    Union,
    overload,
)

import geopandas as gpd

import fused_batch
from fused_batch._global_api import get_api
from fused_batch._str_utils import detect_passing_local_file_as_str, table_to_name
from fused_batch.models._codegen import CustomJobConfig
from fused_batch.models.api import (
    Dataset,
    GeospatialPartitionJobStepConfig,
    JobConfig,
    JobStepConfig,
    JoinJobStepConfig,
    ListDetails,
    MapJobStepConfig,
    NonGeospatialPartitionJobStepConfig,
    Table,
)
from fused_batch.models.api.job import GDALOpenConfig, JoinType, RootAnyJobStepConfig
from fused_batch.models.internal import (
    DatasetInputV2,
    DatasetInputV2Table,
    DatasetInputV2Type,
    DatasetOutputV2,
    Jobs,
)
from fused_batch.models.internal.dataset import RootAnyDatasetInput
from fused_batch.models.schema import Schema
from fused_batch.models.udf import AnyBaseUdf
from fused_batch.warnings import (
    FusedDefaultWarning,
    FusedIgnoredWarning,
    FusedTypeWarning,
)

if TYPE_CHECKING:
    from xyzservices.lib import TileProvider


def _table_to_v2_table(
    table: Union[Table, str], *, read_sidecar: Union[Sequence[str], bool] = False
) -> DatasetInputV2Table:
    table_url = table.url if isinstance(table, Table) else table
    table_name = table_to_name(table_url)
    read_sidecar_bool = read_sidecar is not None and (
        read_sidecar if isinstance(read_sidecar, bool) else (table_name in read_sidecar)
    )
    return DatasetInputV2Table(
        url=table_url,
        read_sidecar_files=read_sidecar_bool,
    )


def _create_table_objs(
    tables: Iterable[Union[Table, str]],
    read_sidecar: Union[Sequence[str], bool] = False,
):
    ret = [_table_to_v2_table(t, read_sidecar=read_sidecar) for t in tables]

    if not isinstance(read_sidecar, bool):
        unapplied_sidecar_table_names: List[str] = []
        applied_sidecar_table_names = set(
            [table_to_name(t.url) for t in ret if t.read_sidecar_files is True]
        )
        for sidecar_table_name in read_sidecar:
            if sidecar_table_name not in applied_sidecar_table_names:
                unapplied_sidecar_table_names.append(sidecar_table_name)

        if unapplied_sidecar_table_names:
            warnings.warn(
                f"Some table names to read sidecars from were not applied: {unapplied_sidecar_table_names}",
                FusedIgnoredWarning,
            )

    return ret


def get_jobs(
    n: int = 5,
    *,
    skip: int = 0,
    per_request: int = 25,
    max_requests: Optional[int] = None,
) -> Jobs:
    """Get the job history.

    Args:
        n: The number of jobs to fetch. Defaults to 5.

    Keyword Args:
        skip: Where in the job history to begin. Defaults to 0, which retrieves the most recent job.
        per_request: Number of jobs per request to fetch. Defaults to 25.
        max_requests: Maximum number of requests to make. May be None to fetch all jobs. Defaults to 1.

    Returns:
        The job history.
    """
    api = get_api()
    return api.get_jobs(
        n=n, skip=skip, per_request=per_request, max_requests=max_requests
    )


def get_udfs(
    n: int = 10,
    *,
    skip: int = 0,
    per_request: int = 25,
    max_requests: Optional[int] = None,
    by: Literal["name", "id", "slug"] = "name",
    whose: Literal["self", "public"] = "self",
):
    """
    Fetches a list of UDFs.

    Args:
        n: The total number of UDFs to fetch. Defaults to 10.
        skip: The number of UDFs to skip before starting to collect the result set. Defaults to 0.
        per_request: The number of UDFs to fetch in each API request. Defaults to 25.
        max_requests: The maximum number of API requests to make.
        by: The attribute by which to sort the UDFs. Can be "name", "id", or "slug". Defaults to "name".
        whose: Specifies whose UDFs to fetch. Can be "self" for the user's own UDFs or "public" for
            UDFs available publicly. Defaults to "self".

    Returns:
        A list of UDFs.

    Examples:
        Fetch UDFs under the user account:
        ```py
        fused.get_udfs()
        ```
    """
    api = get_api()
    return api.get_udfs(
        n=n,
        skip=skip,
        per_request=per_request,
        max_requests=max_requests,
        by=by,
        whose=whose,
    )


def open_table(
    path: Union[str, DatasetOutputV2],
    *,
    fetch_samples: Optional[bool] = None,
) -> Table:
    """Open a Table object given a path to the root of the table

    Args:
        path: The path to the root of the table on remote storage

    Keyword Args:
        fetch_samples: If True, fetch sample on each table when getting dataset metadata.
    Returns:
        A Table object

    Examples:
        ```py
        table = fused.open_table("s3://my_bucket/path/to/dataset/table/")
        ```
    """
    if isinstance(path, DatasetOutputV2):
        path = path.url

    api = get_api()
    return api.open_table(
        path=path,
        fetch_samples=fetch_samples,
    )


def ingest(
    input: Union[str, Sequence[str], Path, gpd.GeoDataFrame],
    output: Optional[str] = None,
    *,
    output_metadata: Optional[str] = None,
    schema: Optional[Schema] = None,
    file_suffix: Optional[str] = None,
    load_columns: Optional[Sequence[str]] = None,
    remove_cols: Optional[Sequence[str]] = None,
    explode_geometries: bool = False,
    drop_out_of_bounds: Optional[bool] = None,
    partitioning_method: Literal["area", "length", "coords", "rows"] = "rows",
    partitioning_maximum_per_file: Union[int, float, None] = None,
    partitioning_maximum_per_chunk: Union[int, float, None] = None,
    partitioning_max_width_ratio: Union[int, float] = 2,
    partitioning_max_height_ratio: Union[int, float] = 2,
    partitioning_force_utm: Literal["file", "chunk", None] = "chunk",
    partitioning_split_method: Literal["mean", "median"] = "mean",
    subdivide_method: Literal["area", None] = None,
    subdivide_start: Optional[float] = None,
    subdivide_stop: Optional[float] = None,
    split_identical_centroids: bool = True,
    target_num_chunks: int = 5000,
    lonlat_cols: Optional[Tuple[str, str]] = None,
    gdal_config: Union[GDALOpenConfig, Dict[str, Any], None] = None,
) -> GeospatialPartitionJobStepConfig:
    """Ingest a dataset into the Fused partitioned format.

    Args:
        input: A GeoPandas `GeoDataFrame` or a path to file or files on S3 to ingest. Files may be Parquet or another geo data format.
        output: Location on S3 to write the `main` table to.
        output_metadata: Location on S3 to write the `fused` table to.
        schema: Schema of the data to be ingested. This is optional and will be inferred from the data if not provided.
        file_suffix: filter which files are used for ingestion. If `input` is a directory on S3, all files under that directory will be listed and used for ingestion. If `file_suffix` is not None, it will be used to filter paths by checking the trailing characters of each filename. E.g. pass `file_suffix=".geojson"` to include only GeoJSON files inside the directory.
        load_columns: Read only this set of columns when ingesting geospatial datasets. Defaults to all columns.
        remove_cols: The named columns to drop when ingesting geospatial datasets. Defaults to not drop any columns.
        explode_geometries: Whether to unpack multipart geometries to single geometries when ingesting geospatial datasets, saving each part as its own row. Defaults to `False`.
        drop_out_of_bounds: Whether to drop geometries outside of the expected WGS84 bounds. Defaults to True.
        partitioning_method: The method to use for grouping rows into partitions. Defaults to `"rows"`.

            - `"area"`: Construct partitions where all contain a maximum total area among geometries.
            - `"length"`: Construct partitions where all contain a maximum total length among geometries.
            - `"coords"`: Construct partitions where all contain a maximum total number of coordinates among geometries.
            - `"rows"`: Construct partitions where all contain a maximum number of rows.

        partitioning_maximum_per_file: Maximum value for `partitioning_method` to use per file. If `None`, defaults to _1/10th_ of the total value of `partitioning_method`. So if the value is `None` and `partitioning_method` is `"area"`, then each file will be have no more than 1/10th the total area of all geometries. Defaults to `None`.
        partitioning_maximum_per_chunk: Maximum value for `partitioning_method` to use per chunk. If `None`, defaults to _1/100th_ of the total value of `partitioning_method`. So if the value is `None` and `partitioning_method` is `"area"`, then each file will be have no more than 1/100th the total area of all geometries. Defaults to `None`.
        partitioning_max_width_ratio: The maximum ratio of width to height of each partition to use in the ingestion process. So for example, if the value is `2`, then if the width divided by the height is greater than `2`, the box will be split in half along the horizontal axis. Defaults to `2`.
        partitioning_max_height_ratio: The maximum ratio of height to width of each partition to use in the ingestion process. So for example, if the value is `2`, then if the height divided by the width is greater than `2`, the box will be split in half along the vertical axis. Defaults to `2`.
        partitioning_force_utm: Whether to force partitioning within UTM zones. If set to `"file"`, this will ensure that the centroid of all geometries per _file_ are contained in the same UTM zone. If set to `"chunk"`, this will ensure that the centroid of all geometries per _chunk_ are contained in the same UTM zone. If set to `None`, then no UTM-based partitioning will be done. Defaults to "chunk".
        partitioning_split_method: How to split one partition into children. Defaults to `"mean"` (this may change in the future).

            - `"mean"`: Split each axis according to the mean of the centroid values.
            - `"median"`: Split each axis according to the median of the centroid values.

        subdivide_method: The method to use for subdividing large geometries into multiple rows. Currently the only option is `"area"`, where geometries will be subdivided based on their area (in WGS84 degrees).
        subdivide_start: The value above which geometries will be subdivided into smaller parts, according to `subdivide_method`.
        subdivide_stop: The value below which geometries will not be subdivided into smaller parts, according to `subdivide_method`. Recommended to be equal to subdivide_start. If `None`, geometries will be subdivided up to a recursion depth of 100 or until the subdivided geometry is rectangular.
        split_identical_centroids: If `True`, should split a partition that has
            identical centroids (such as if all geometries in the partition are the
            same) if there are more such rows than defined in "partitioning_maximum_per_file" and
            "partitioning_maximum_per_chunk".
        target_num_chunks: The target for the number of files if `partitioning_maximum_per_file` is None. Note that this number is only a _target_ and the actual number of files generated can be higher or lower than this number, depending on the spatial distribution of the data itself.
        lonlat_cols: Names of longitude, latitude columns to construct point geometries from.

            If your point columns are named `"x"` and `"y"`, then pass:

            ```py
            fused.ingest(
                ...,
                lonlat_cols=("x", "y")
            )
            ```

            This only applies to reading from Parquet files. For reading from CSV files, pass options to `gdal_config`.

        gdal_config: Configuration options to pass to GDAL for how to read these files. For all files other than Parquet files, Fused uses GDAL as a step in the ingestion process. For some inputs, like CSV files or zipped shapefiles, you may need to pass some parameters to GDAL to tell it how to open your files.

            This config is expected to be a dictionary with up to two keys:

            - `layer`: `str`. Define the layer of the input file you wish to read when the source contains multiple layers, as in GeoPackage.
            - `open_options`: `Dict[str, str]`. Pass in key-value pairs with GDAL open options. These are defined on each driver's page in the GDAL documentation. For example, the [CSV driver](https://gdal.org/drivers/vector/csv.html) defines [these open options](https://gdal.org/drivers/vector/csv.html#open-options) you can pass in.

            For example, if you're ingesting a CSV file with two columns
            `"longitude"` and `"latitude"` denoting the coordinate information, pass

            ```py
            fused.ingest(
                ...,
                gdal_config={
                    "open_options": {
                        "X_POSSIBLE_NAMES": "longitude",
                        "Y_POSSIBLE_NAMES": "latitude",
                    }
                }
            )
            ```
    Returns:

        Configuration object describing the ingestion process. Call `.execute` on this object to start a job.


    Examples:
        For example, to ingest the California Census dataset for the year 2022:
        ```py
        job = fused.ingest(
            input="https://www2.census.gov/geo/tiger/TIGER_RD18/STATE/06_CALIFORNIA/06/tl_rd22_06_bg.zip",
            output="s3://fused-sample/census/ca_bg_2022/main/",
            output_metadata="s3://fused-sample/census/ca_bg_2022/fused/",
            explode_geometries=True,
            partitioning_maximum_per_file=2000,
            partitioning_maximum_per_chunk=200,
        ).execute()
        ```
    """
    remove_cols = remove_cols if remove_cols else []
    if (
        subdivide_start is not None or subdivide_stop is not None
    ) and subdivide_method is None:
        raise ValueError(
            'subdivide_start or subdivide_stop require subdivide_method be specified (it should be "area")'
        )
    api = get_api()
    input = detect_passing_local_file_as_str(input)
    input = api._replace_df_input(input)
    return GeospatialPartitionJobStepConfig(
        input=input,
        output=output,
        output_metadata=output_metadata,
        table_schema=schema,
        file_suffix=file_suffix,
        load_columns=load_columns,
        remove_cols=remove_cols,
        explode_geometries=explode_geometries,
        drop_out_of_bounds=drop_out_of_bounds,
        lonlat_cols=lonlat_cols,
        partitioning_maximum_per_file=partitioning_maximum_per_file,
        partitioning_maximum_per_chunk=partitioning_maximum_per_chunk,
        partitioning_max_width_ratio=partitioning_max_width_ratio,
        partitioning_max_height_ratio=partitioning_max_height_ratio,
        partitioning_method=partitioning_method,
        partitioning_force_utm=partitioning_force_utm,
        partitioning_split_method=partitioning_split_method,
        subdivide_start=subdivide_start,
        subdivide_stop=subdivide_stop,
        subdivide_method=subdivide_method,
        split_identical_centroids=split_identical_centroids,
        target_num_chunks=target_num_chunks,
        gdal_config=GDALOpenConfig() if gdal_config is None else gdal_config,
    )


def ingest_nongeospatial(
    input: Union[str, Sequence[str], Path, gpd.GeoDataFrame],
    output: Optional[str] = None,
    *,
    output_metadata: Optional[str] = None,
    partition_col: Optional[str] = None,
    partitioning_maximum_per_file: int = 2_500_000,
    partitioning_maximum_per_chunk: int = 65_000,
) -> NonGeospatialPartitionJobStepConfig:
    """Ingest a dataset into the Fused partitioned format.

    Args:
        input: A GeoPandas `GeoDataFrame` or a path to file or files on S3 to ingest. Files may be Parquet or another geo data format.
        output: Location on S3 to write the `main` table to.
        output_metadata: Location on S3 to write the `fused` table to.
        partition_col: Partition along this column for nongeospatial datasets.
        partitioning_maximum_per_file: Maximum number of items to store in a single file. Defaults to 2,500,000.
        partitioning_maximum_per_chunk: Maximum number of items to store in a single file. Defaults to 65,000.

    Returns:

        Configuration object describing the ingestion process. Call `.execute` on this object to start a job.

    Examples:
        ```py
        job = fused.ingest_nongeospatial(
            input=gdf,
            output="s3://sample-bucket/file.parquet",
        ).execute()
        ```
    """
    api = get_api()
    input = detect_passing_local_file_as_str(input)
    input = api._replace_df_input(input)
    return NonGeospatialPartitionJobStepConfig(
        input=input,
        output=output,
        output_metadata=output_metadata,
        partition_col=partition_col,
        partitioning_maximum_per_file=partitioning_maximum_per_file,
        partitioning_maximum_per_chunk=partitioning_maximum_per_chunk,
    )


def map(
    dataset: Union[str, Dataset, Table],
    output_table: Optional[str] = None,
    udf: Union[AnyBaseUdf, None] = None,
    *,
    cache_locally: bool = False,
) -> MapJobStepConfig:
    """Construct a `map` config from this Table

    Args:
        output_table: Where to save the output of this operation. Defaults to `None`, which will not save the output.
        udf: A user-defined function to run in this map. Defaults to None.

    Keyword Args:
        tables: The attribute tables to include in the map reduce. Defaults to ("main",).
        cache_locally: Advanced: whether to cache all the partitions locally in the map job. Defaults to False.

    Returns:
        An object describing the map configuration.
    """
    api = get_api()
    if isinstance(dataset, str):
        dataset = api.open_table(dataset, fetch_samples=False)

    return dataset.map(
        output_table=output_table,
        udf=udf,
        cache_locally=cache_locally,
    )


def join(
    dataset: Union[str, Dataset, Table],
    other: Union[str, Dataset, Table],
    output_table: Optional[str] = None,
    udf: Union[AnyBaseUdf, None] = None,
    *,
    how: Union[JoinType, Literal["left", "inner"]] = "inner",
    left_cache_locally: bool = False,
    right_cache_locally: bool = False,
    buffer_distance: Optional[float] = None,
) -> JoinJobStepConfig:
    """Construct a join config from two tables

    Args:
        other: The other Dataset object to join on
        output_table: Where to save the output of this operation. Defaults to `None`, which will not save the output.
        udf: The user-defined function to run in the join

    Keyword Args:
        how: The manner of doing the join. Currently  Defaults to "inner".
        left_cache_locally: Whether to cache the left dataset locally in the join. Defaults to False.
        right_cache_locally: Whether to cache the right dataset locally in the join. Defaults to False.
        buffer_distance: The size of the buffer (in meters) on the left table to use during the join. Defaults to None.

    Returns:
        An object describing the join configuration.

    Examples:
        ```py
        import fused

        left_table = fused.open_table("s3://bucket/path/to/table")
        other_table = fused.open_table("s3://bucket/path/to/table")
        join_config = fused.join(left_table, other_table)
        ```


    """
    api = get_api()

    if isinstance(dataset, str):
        dataset = api.open_table(dataset, fetch_samples=False)

    if isinstance(other, str):
        other = api.open_table(other, fetch_samples=False)

    return dataset.join(
        other=other,
        output_table=output_table,
        udf=udf,
        how=how,
        left_cache_locally=left_cache_locally,
        right_cache_locally=right_cache_locally,
        buffer_distance=buffer_distance,
    )


def job(
    input: Union[
        str,
        Dict,
        JobStepConfig,
        JobConfig,
        Sequence[Union[Dict, JobStepConfig, JobConfig]],
    ],
    content_type: Optional[str] = None,
    ignore_chunk_error: bool = False,
) -> JobConfig:
    """Construct a JobConfig

    Args:
        input: A object or list of objects describing job steps.
        content_type: How to interpret `input` when it is a string. E.g. "json" for JSON or "fused_job_id" for a Fused Job ID.

    Returns:
        A combined job config.
    """
    if isinstance(input, str):
        if content_type is None:
            if _is_valid_uuid(input):
                content_type = "fused_job_id"
            else:
                warnings.warn(
                    'content_type is not set, assuming "json"', FusedDefaultWarning
                )
                content_type = "json"
        if content_type == "fused_job_id":
            api = get_api()
            return api.get_job_config(input)
        if content_type == "json":
            return JobConfig(
                steps=[RootAnyJobStepConfig.model_validate_json(input).root]
            )
        else:
            assert (
                False
            ), f"Unknown content type: {content_type}. Should be one of `json` or `fused_job_id`."
    elif isinstance(input, dict):
        return JobConfig(steps=[RootAnyJobStepConfig.model_validate(input).root])
    elif isinstance(input, JobStepConfig):
        config = JobConfig(steps=[input])
    elif isinstance(input, JobConfig):
        return job(input.model_copy(deep=True).steps)
    else:
        # Assumed to be sequence
        steps: List[JobStepConfig] = []
        for input_step in input:
            if isinstance(input_step, dict):
                step = RootAnyJobStepConfig.model_validate(input_step).root
                steps.append(step)
            elif isinstance(input_step, JobStepConfig):
                steps.append(input_step.model_copy(deep=True))
            elif isinstance(input_step, JobConfig):
                steps.extend(input_step.model_copy(deep=True).steps)
            else:
                assert False, "expected dict or JobStepConfig"

        config = JobConfig(steps=steps, ignore_chunk_error=ignore_chunk_error)

    return config


def _whoami():
    """
    Returns information on the currently logged in user
    """
    api = get_api()
    return api._whoami()


def plot(
    gdf: gpd.GeoDataFrame,
    source: Union[str, "TileProvider", Ellipsis] = ...,
    **geopandas_kwargs: Dict[str, Any],
):
    """
    Plot a GeoDataFrame on a map using contextily to add basemap

    Args:
        gdf: A GeoPandas `GeoDataFrame` to plot.
        source : Basemap to use. Accepts an xyzservices.TileProvider object or str.
            [Optional. Default: CartoDB DarkMatterNoLabels]
            The tile source: web tile provider, a valid input for a query of a
            :class:`xyzservices.TileProvider` by a name from ``xyzservices.providers`` or
            path to local file.
        **geopandas_kwargs: Additional keyword arguments to pass to `gdf.plot`.
    """
    import contextily as cx

    if source is Ellipsis:
        source = cx.providers.CartoDB.DarkMatterNoLabels

    ax = gdf.plot(**geopandas_kwargs)
    if hasattr(gdf, "crs") and gdf.crs is not None:
        crs = gdf.crs.to_string()
    else:
        crs = "WGS84"
        warnings.warn("CRS not specified, defaulting to WGS84", FusedDefaultWarning)
    return cx.add_basemap(ax, crs=crs, source=source)


def delete(
    path: str,
    max_deletion_depth: Union[int, Literal["unlimited"]] = 2,
) -> bool:
    """Delete the files at the path.

    Args:
        path: Directory or file to delete, like `fd://my-old-table/`
        max_deletion_depth: If set (defaults to 2), the maximum depth the operation will recurse to.
                            This option is to help avoid accidentally deleting more data that intended.
                            Pass `"unlimited"` for unlimited.


    Examples:
        ```python
        fused.delete("fd://bucket-name/deprecated_table/")
        ```
    """
    api = get_api()
    return api.delete(path, max_deletion_depth=max_deletion_depth)


@overload
def list(path: str, *, details: Literal[False] = False) -> List[str]:
    ...


@overload
def list(path: str, *, details: Literal[True]) -> List[ListDetails]:
    ...


def list(path: str, *, details: bool = False):
    """List the files at the path.

    Args:
        path: Parent directory URL, like `fd://bucket-name/`

    Keyword Args:
        details: If True, return additional metadata about each record.

    Returns:
        A list of paths as URLs, or as metadata objects.

    Examples:
        ```python
        fused.list("fd://bucket-name/")
        ```
    """
    api = get_api()
    return api.list(path, details=details)


def get(path: str) -> bytes:
    """Download the contents at the path to memory.

    Args:
        path: URL to a file, like `fd://bucket-name/file.parquet`

    Returns:
        bytes of the file

    Examples:
        ```python
        fused.get("fd://bucket-name/file.parquet")
        ```
    """
    api = get_api()
    return api.get(path)


def download(path: str, local_path: Union[str, Path]) -> None:
    """Download the contents at the path to disk.

    Args:
        path: URL to a file, like `fd://bucket-name/file.parquet`
        local_path: Path to a local file.
    """
    api = get_api()
    api.download(path, local_path=local_path)


def upload(local_path: Union[str, Path, bytes, BinaryIO], remote_path: str) -> None:
    """Upload local file to S3.

    Args:
        local_path: Either a path to a local file (`str`, `Path`) or the contents to upload.
                    Any string will be treated as a Path, if you wish to upload the contents of
                    the string, first encode it: `s.encode("utf-8")`
        remote_path: URL to upload to, like `fd://new-file.txt`

    Examples:
        To upload a local json file to your Fused-managed S3 bucket:
        ```py
        fused.upload("my_file.json", "fd://my_bucket/my_file.json")
        ```
    """
    api = get_api()
    if isinstance(local_path, str):
        # We assume any string being passed in is a path, rather than the contents
        # to upload.
        local_path = Path(local_path)
        if not local_path.exists():
            warnings.warn(
                '`local_path` is being treated as a path but it does not exist. If you wish to upload the contents of the string, encode it to bytes first with `.encode("utf-8")',
                FusedTypeWarning,
            )
    if isinstance(local_path, Path):
        local_path = local_path.read_bytes()
    api.upload(path=remote_path, data=local_path)


def sign_url(path: str) -> str:
    """Create a signed URL to access the path.

    This function may not check that the file represented by the path exists.

    Args:
        path: URL to a file, like `fd://bucket-name/file.parquet`

    Returns:
        HTTPS URL to access the file using signed access.

    Examples:
        ```python
        fused.sign_url("fd://bucket-name/table_directory/file.parquet")
        ```
    """
    api = get_api()
    return api.sign_url(path)


def sign_url_prefix(path: str) -> Dict[str, str]:
    """Create signed URLs to access all blobs under the path.

    Args:
        path: URL to a prefix, like `fd://bucket-name/some_directory/`

    Returns:
        Dictionary mapping from blob store key to signed HTTPS URL.

    Examples:
        ```python
        fused.sign_url_prefix("fd://bucket-name/table_directory/")
        ```
    """
    api = get_api()
    return api.sign_url_prefix(path)


def zip_tables(
    tables: Iterable[Union[Table, str]],
    *,
    read_sidecar: Union[Sequence[str], bool] = False,
) -> DatasetInputV2:
    """Create a job input that zips the columns of tables together. This takes the partitions from all the listed tables and combines them (as new columns) into a single DataFrame per chunk.

    Args:
        tables: The tables to zip together

    Keyword Args:
        read_sidecar: Whether to read sidecar information, either a sequence of table names (i.e. the last part of the table path) to read it from or a boolean which will be applied to all tables (default False).
    """
    return DatasetInputV2(
        tables=_create_table_objs(tables=tables, read_sidecar=read_sidecar),
        operation=DatasetInputV2Type.ZIP,
    )


def union_tables(
    tables: Iterable[Union[Table, str]],
    *,
    read_sidecar: Union[Sequence[str], bool] = False,
) -> DatasetInputV2:
    """Create a job input that unions the partitions of tables together. This takes the partitions from all the listed tables (which should have the same schema) and runs the operation over each partition.

    Args:
        tables: The tables to union together

    Keyword Args:
        read_sidecar: Whether to read sidecar information, either a sequence of table names (i.e. the last part of the table path) to read it from or a boolean which will be applied to all tables (default False).
    """
    return DatasetInputV2(
        tables=_create_table_objs(tables=tables, read_sidecar=read_sidecar),
        operation=DatasetInputV2Type.UNION,
    )


def load_job(path: Union[str, Path, BinaryIO]):
    @contextmanager
    def change_directory(path):
        if isinstance(path, (str, Path)):
            path_obj = Path(path)
        else:
            path_obj = path
        # Store the current working directory
        old_directory = os.getcwd()

        try:
            if not isinstance(path_obj, Path) or path_obj.is_file():
                # If .zip, create a temporary directory for zip.
                with tempfile.TemporaryDirectory() as temp_dir:
                    # Extract the contents of the zip file.
                    with zipfile.ZipFile(path, "r") as zip_ref:
                        zip_ref.extractall(temp_dir)

                    os.chdir(temp_dir)
                    yield temp_dir

            else:
                os.chdir(path)
                yield
        finally:
            # Revert back to the original directory.
            os.chdir(old_directory)

    with change_directory(path):
        # Open and read the JSON file
        with Path("meta.json").open("r") as file:
            try:
                data = json.load(file)
            except json.JSONDecodeError as e:
                print(f"Error decoding JSON: {e}")

        # Validate the asset JSON and apply necessary migrations
        validated_job = CustomJobConfig(**data["job_config"])

        jobs = []
        for step in validated_job.steps:
            local_headers = [
                header["module_name"] + ".py" for header in step["udf"]["headers"]
            ]

            # TODO: Add type hint "List[DatasetInputBase]"
            # TODO: Switch to DatasetInputBase.parse_obj below
            input_params = []
            arg_list: Optional[Any] = None
            join_is_singlefile: Optional[bool] = None
            buffer_distance: Optional[float] = None
            join_how: Optional[str] = None
            if step["type"] == "join":
                input_params = [
                    RootAnyDatasetInput.model_validate(step["input_left"]).root,
                    RootAnyDatasetInput.model_validate(step["input_right"]).root,
                ]
                buffer_distance = step.get("buffer_distance")
                join_how = step.get("how")
            elif step["type"] == "map":
                input_params = [RootAnyDatasetInput.model_validate(step["input"]).root]
            elif step["type"] == "join_singlefile":
                input_params = [
                    RootAnyDatasetInput.model_validate(step["input_left"]).root,
                    step["input_right"],
                ]
                join_is_singlefile = True
            elif step["type"] == "udf":
                arg_list = step.get("input")
            udf_registry = fused_batch.experimental.load_udf(
                udf_paths=[step["udf"]["source"]], header_paths=local_headers
            )
            # TODO: Note this may misdetect if you try to read more than one UDF with the same entrypoint name
            detected_udf_name = (
                step["udf"]["name"]
                if step["udf"]["name"] in udf_registry
                else step["udf"]["entrypoint"]
            )
            # Materialize job with input parameters
            # Note this code is expected to not reload the output of the job
            job = udf_registry[detected_udf_name](
                *input_params,
                **step["udf"]["parameters"],
                buffer_distance=buffer_distance,
                join_is_singlefile=join_is_singlefile,
                **({"join_how": join_how} if join_how is not None else {}),
                arg_list=arg_list,
            )

            job.udf.metadata = step["udf"].get("metadata")
            job.metadata = step.get("metadata")

            # Materialize job and append
            jobs.append(job)

        job = fused_batch.experimental.job(jobs)
        job.metadata = (
            validated_job.metadata
        )  # TODO: job.export does not write this metadata

    assert isinstance(job, JobConfig)

    return job


def _is_valid_uuid(s: str) -> bool:
    try:
        uuid.UUID(s)
        return True
    except ValueError:
        return False



from pathlib import Path
from typing import Any, Union

import requests

from fused.models.api.udf_access_token import is_udf_token

from .core import load_udf_from_fused, load_udf_from_github, load_udf_from_shared_token
from .core._impl._reimports import AnyBaseUdf


def load(url_or_udf: Union[str, Path], /, *, cache_key: Any = None) -> AnyBaseUdf:
    """
    Loads a UDF from various sources including GitHub URLs,
    and a Fused platform-specific identifier.

    This function supports loading UDFs from a GitHub repository URL, or a Fused
    platform-specific identifier composed of an email and UDF name. It intelligently
    determines the source type based on the format of the input and retrieves the UDF
    accordingly.

    Args:
        url_or_udf: A string or Path object representing the location of the UDF. This can be
            a GitHub URL starting with "https://github.com", or a Fused platform-specific
            identifier in the format "email/udf_name".
        cache_key: An optional key used for caching the loaded UDF. If provided, the function
            will attempt to load the UDF from cache using this key before attempting to
            load it from the specified source. Defaults to None, indicating no caching.

    Returns:
        AnyBaseUdf: An instance of the loaded UDF.

    Raises:
        ValueError: If the URL or Fused platform-specific identifier format is incorrect or
            cannot be parsed.
        Exception: For errors related to network issues, file access permissions, or other
            unforeseen errors during the loading process.

    Examples:
        Load a UDF from a GitHub URL:
        ```py
        udf = fused.load("https://github.com/fusedio/udfs/tree/main/public/REM_with_HyRiver/")
        ```

        Load a UDF using a Fused platform-specific identifier:
        ```py
        udf = fused.load("username@fused.io/REM_with_HyRiver")
        ```
    """
    if is_udf_token(url_or_udf):
        if url_or_udf.startswith("UDF_"):
            url_or_udf = f"public@fused.io/{url_or_udf.removeprefix('UDF_')}"
        else:
            try:
                return load_udf_from_shared_token(url_or_udf)
            except requests.HTTPError:
                raise ValueError(
                    "It looks like you tried to load a UDF from a shared token "
                    "without having access to the source code. "
                    "You may want to call `run` instead of `load`."
                )

    if url_or_udf.startswith("https://github.com"):
        return load_udf_from_github(url_or_udf, cache_key=cache_key)
    else:
        parts = url_or_udf.split("/", maxsplit=1)
        if len(parts) == 2:
            email_or_handle, udf_name = parts
            return load_udf_from_fused(email_or_handle, udf_name, cache_key=cache_key)
        elif len(parts) == 1:
            udf_name = parts[0]
            return load_udf_from_fused(udf_name, cache_key=cache_key)



"""A module for 🪄🧙 magic 🪄🧙"""
# All IPython-dependent code should be in this folder

# ruff: noqa: F401

import pandas

from .._udf.context import context
from .output import output

try:
    import IPython

    from .magics import load_ipython_extension

    ipython = IPython.get_ipython()

    if ipython is None:
        raise ImportError("Not running inside IPython.")

    # automatically load the %%fused magic into the environment
    # This takes the place of %load_ext fused
    load_ipython_extension(ipython)
except ImportError:
    pass  # Not in IPython

# We set pandas' 2.0 copy_on_write to ON by default to avoid pesky
# SettingWithCopyWarningErrors
pandas.options.mode.copy_on_write = True



from __future__ import annotations

import shutil
import time
import uuid
import warnings
from functools import lru_cache
from io import SEEK_SET
from pathlib import Path
from tempfile import TemporaryFile
from typing import (
    Any,
    BinaryIO,
    Dict,
    List,
    Literal,
    Optional,
    Sequence,
    Union,
    overload,
)

import geopandas as gpd
import pandas as pd
import requests

import fused_batch
import fused_batch.models.request as request_models
from fused_batch._auth import AUTHORIZATION
from fused_batch._deserialize import (
    zip_to_join_args,
    zip_to_join_singlefile_args,
    zip_to_map_args,
)
from fused_batch._deserialize_parquet import parquet_to_df
from fused_batch._global_api import set_api, set_api_class
from fused_batch._options import PROD_DEFAULT_BASE_URL
from fused_batch._options import options as OPTIONS
from fused_batch._request import raise_for_status
from fused_batch.api._open_dataset import post_open_table
from fused_batch.models import JoinInput, JoinSingleFileInput, MapInput
from fused_batch.models.api import (
    JobConfig,
    JobStepConfig,
    JoinJobStepConfig,
    JoinSinglefileJobStepConfig,
    ListDetails,
    MapJobStepConfig,
    Table,
    UdfAccessToken,
    UdfAccessTokenList,
)
from fused_batch.models.api._folder import Folder
from fused_batch.models.internal import Jobs, RunResponse
from fused_batch.models.internal.job import CoerceableToJobId, _object_to_job_id
from fused_batch.models.request import WHITELISTED_INSTANCE_TYPES
from fused_batch.models.udf._udf_registry import UdfRegistry
from fused_batch.models.udf.base_udf import (
    METADATA_FUSED_EXPLORER_TAB,
    METADATA_FUSED_ID,
    METADATA_FUSED_SLUG,
)
from fused_batch.models.udf.udf import AnyBaseUdf, RootAnyBaseUdf
from fused_batch.warnings import FusedNonProductionWarning


class FusedAPI:
    """API for running jobs in the Fused service."""

    base_url: str

    def __init__(
        self,
        *,
        base_url: Optional[str] = None,
        set_global_api: bool = True,
        credentials_needed: bool = True,
    ):
        """Create a FusedAPI instance.

        Keyword Args:
            base_url: The Fused instance to send requests to. Defaults to `https://www.fused.io/server/v1`.
            set_global_api: Set this as the global API object. Defaults to True.
            credentials_needed: If True, automatically attempt to log in. Defaults to True.
        """
        if credentials_needed:
            AUTHORIZATION.initialize()
        base_url = base_url or OPTIONS.base_url

        self.base_url = base_url
        self._check_is_prod()

        if set_global_api:
            set_api(self)

    def _check_is_prod(self):
        if self.base_url != PROD_DEFAULT_BASE_URL:
            warnings.warn(
                "FusedAPI is connected to non-production", FusedNonProductionWarning
            )

    def sample_map(
        self,
        config: MapJobStepConfig,
        *,
        file_id: Optional[Union[str, int]] = None,
        chunk_id: Optional[int] = None,
        n_rows: Optional[int] = None,
    ) -> MapInput:
        """Fetch a sample of an operation

        Args:
            config: The configuration to sample from.

        Keyword Args:
            file_id: The identifier of this file. Defaults to None.
            chunk_id: The numeric index of the chunk within the file to fetch. Defaults to None.
            n_rows: The maximum number of rows to sample. Defaults to None for all rows in the chunk.
        """
        assert isinstance(config, MapJobStepConfig)
        url = f"{self.base_url}/sample_args/map"

        body: Dict[str, Any] = config.model_dump()
        params = request_models.SampleMapRequest(
            file_id=file_id, chunk_id=chunk_id, n_rows=n_rows
        )

        self._check_is_prod()
        r = requests.post(
            url=url,
            params=params.model_dump(),
            json=body,
            headers=self._generate_headers(),
            timeout=OPTIONS.request_timeout,
        )
        raise_for_status(r)
        return zip_to_map_args(r.content, n_rows=n_rows)

    def _whole_file_sample_map(
        self,
        config: MapJobStepConfig,
        *,
        file_id: Union[str, int],
        n_rows: Optional[int] = None,
    ) -> MapInput:
        assert isinstance(config, MapJobStepConfig)
        url = f"{self.base_url}/sample_args/whole_file"

        body: Dict[str, Any] = config.model_dump()
        params = request_models.WholeFileSampleMapRequest(
            file_id=file_id, n_rows=n_rows
        )

        self._check_is_prod()
        r = requests.post(
            url=url,
            params=params.model_dump(),
            json=body,
            headers=self._generate_headers(),
            timeout=OPTIONS.request_timeout,
        )
        raise_for_status(r)
        return zip_to_map_args(r.content, n_rows=n_rows)

    def download_table_bbox(
        self,
        path: str,
        minx: float,
        miny: float,
        maxx: float,
        maxy: float,
        n_rows: Optional[int] = None,
        columns: Optional[List[str]] = None,
        clip: bool = True,
        buffer: Optional[float] = None,
    ) -> Union[pd.DataFrame, gpd.GeoDataFrame]:
        url = f"{self.base_url}/table/download/bbox"

        params = request_models.GetTableBboxRequest(
            path=path,
            bbox_minx=minx,
            bbox_miny=miny,
            bbox_maxx=maxx,
            bbox_maxy=maxy,
            n_rows=n_rows,
            columns=columns,
            clip=clip,
            buffer=buffer,
        )

        self._check_is_prod()
        r = requests.get(
            url=url,
            params=params.model_dump(),
            headers=self._generate_headers(),
            timeout=OPTIONS.request_timeout,
        )
        raise_for_status(r)
        return parquet_to_df(r.content)

    def sample_join(
        self,
        config: JoinJobStepConfig,
        *,
        file_id: Optional[Union[str, int]] = None,
        chunk_id: Optional[int] = None,
        n_rows: Optional[int] = None,
    ) -> JoinInput:
        """Fetch a sample of an operation

        Args:
            config: The configuration to sample from.

        Keyword Args:
            file_id: The identifier of this file. Defaults to None.
            chunk_id: The numeric index of the chunk within the file to fetch. Defaults to None.
            n_rows: The maximum number of rows to sample from the left dataset. Defaults to None for all rows in the chunk.
        """
        url = f"{self.base_url}/sample_args/join"
        body: Dict[str, Any] = config.model_dump()

        params = request_models.SampleJoinRequest(
            file_id=file_id, chunk_id=chunk_id, n_rows=n_rows
        )

        self._check_is_prod()
        r = requests.post(
            url=url,
            params=params.model_dump(),
            json=body,
            headers=self._generate_headers(),
            timeout=OPTIONS.request_timeout,
        )
        raise_for_status(r)
        return zip_to_join_args(r.content, n_rows=n_rows)

    def sample_single_file_join(
        self,
        config: JoinSinglefileJobStepConfig,
        *,
        file_id: Optional[Union[str, int]] = None,
        chunk_id: Optional[int] = None,
        n_rows: Optional[int] = None,
    ) -> JoinSingleFileInput:
        """Fetch a sample of an operation

        Args:
            config: The configuration to sample from.

        Keyword Args:
            file_id: The identifier of this file. Defaults to None.
            chunk_id: The numeric index of the chunk within the file to fetch. Defaults to None.
            n_rows: The maximum number of rows to sample from the left dataset. Defaults to None for all rows in the chunk.
        """
        url = f"{self.base_url}/sample_args/join_singlefile"
        body: Dict[str, Any] = config.model_dump()

        params = request_models.SampleSingleFileJoinRequest(
            file_id=file_id, chunk_id=chunk_id, n_rows=n_rows
        )

        self._check_is_prod()
        r = requests.post(
            url=url,
            params=params.model_dump(),
            json=body,
            headers=self._generate_headers(),
            timeout=OPTIONS.request_timeout,
        )
        raise_for_status(r)
        return zip_to_join_singlefile_args(r.content, n_rows=n_rows)

    def start_job(
        self,
        config: Union[JobConfig, JobStepConfig],
        *,
        instance_type: Optional[WHITELISTED_INSTANCE_TYPES] = None,
        region: Optional[str] = None,
        disk_size_gb: Optional[int] = None,
        additional_env: Optional[Sequence[str]] = ("FUSED_CREDENTIAL_PROVIDER=ec2",),
        image_name: Optional[str] = None,
    ) -> RunResponse:
        """Execute an operation

        Args:
            config: the configuration object to run in the job.

        Keyword Args:
            instance_type: The AWS EC2 instance type to use for the job. Acceptable strings are "m5.large", "m5.xlarge", "m5.2xlarge", "m5.4xlarge", "m5.8xlarge", "m5.12xlarge", "m5.16xlarge", "r5.large", "r5.xlarge", "r5.2xlarge", "r5.4xlarge", "r5.8xlarge", "r5.12xlarge", or "r5.16xlarge". Defaults to None.
            region: The AWS region in which to run. Defaults to None.
            disk_size_gb: The disk size to specify for the job. Defaults to None.
            additional_env: Any additional environment variables to be passed into the job, each in the form KEY=value. Defaults to None.
            image_name: Custom image name to run. Defaults to None for default image.
        """
        url = f"{self.base_url}/run"

        if isinstance(config, JobStepConfig):
            config = JobConfig(steps=[config])

        body = {"config": config.model_dump()}
        if additional_env:
            body["additional_env"] = additional_env
        if image_name:
            body["image_name"] = image_name

        params = request_models.StartJobRequest(
            region=region,
            instance_type=instance_type,
            disk_size_gb=disk_size_gb,
        )

        self._check_is_prod()
        r = requests.post(
            url=url,
            params=params.model_dump(),
            json=body,
            headers=self._generate_headers(),
            timeout=OPTIONS.request_timeout,
        )
        raise_for_status(r)
        return RunResponse.model_validate_json(r.content)

    def save_udf(
        self,
        udf: AnyBaseUdf,
        slug: Optional[str] = None,
        id: Optional[str] = None,
        allow_public_read: bool = False,
        allow_public_list: bool = False,
    ) -> UdfRegistry:
        url = f"{self.base_url}/udf/by-id/{id}" if id else f"{self.base_url}/udf/new"
        body = request_models.SaveUdfRequest(
            slug=slug,
            udf_body=udf.model_dump_json(),
            udf_type=request_models.UdfType.auto,
            allow_public_read=allow_public_read,
            allow_public_list=allow_public_list,
        )

        self._check_is_prod()
        r = requests.post(
            url=url,
            json=body.model_dump(),
            headers=self._generate_headers(),
            timeout=OPTIONS.request_timeout,
        )
        raise_for_status(r)
        # TODO: body
        return r.json()

    def delete_saved_udf(self, id: str):
        url = f"{self.base_url}/udf/by-id/{id}"

        self._check_is_prod()
        r = requests.delete(
            url=url,
            headers=self._generate_headers(),
            timeout=OPTIONS.request_timeout,
        )
        raise_for_status(r)
        # TODO: body
        return r.json()

    def _get_udf(
        self,
        email_or_id: str,
        id: Optional[str] = None,
    ):
        if id is None:
            email_or_id = id
            email_or_id = self._whoami()["email"]

        url = f"{self.base_url}/udf/by-user-email/{email_or_id}/by-slug/{id}"

        r = requests.get(
            url=url,
            headers=self._generate_headers(),
            timeout=OPTIONS.request_timeout,
        )
        raise_for_status(r)
        return r.json()

    def _get_public_udf(
        self,
        id: str,
    ):
        url = f"{self.base_url}/udf/public/by-slug/{id}"

        r = requests.get(
            url=url,
            headers=self._generate_headers(),
            timeout=OPTIONS.request_timeout,
        )
        raise_for_status(r)
        return r.json()

    def _get_code_by_url(self, url: str):
        req_url = f"{self.base_url}/code-proxy/by-url"

        r = requests.get(
            url=req_url,
            params={
                "url": url,
            },
            headers=self._generate_headers(credentials_needed=False),
            timeout=OPTIONS.request_timeout,
        )
        raise_for_status(r)
        return r.json()

    def get_udfs(
        self,
        n: int = 5,
        *,
        skip: int = 0,
        per_request: int = 25,
        max_requests: Optional[int] = 1,
        by: Literal["name", "id", "slug"] = "name",
        whose: Literal["self", "public", "community", "team"] = "self",
    ):
        request_count = 0
        has_content = True
        udfs = []

        assert per_request >= 0

        while has_content:
            if whose == "self":
                url = f"{self.base_url}/udf/self"
            elif whose == "public" or whose == "community":
                url = f"{self.base_url}/udf/public"
            elif whose == "team":
                url = f"{self.base_url}/udf/exec-env/self"
            else:
                raise ValueError(
                    'Invalid value for `whose`, should be one of: "self", "public", "community", "team"'
                )

            params = request_models.ListUdfsRequest(
                skip=skip,
                limit=per_request,
            )
            skip += per_request

            r = requests.get(
                url=url,
                params=params.model_dump(),
                headers=self._generate_headers(),
                timeout=OPTIONS.request_timeout,
            )
            raise_for_status(r)
            udfs_this_request = r.json()
            if udfs_this_request:
                udfs.extend(udfs_this_request)
            else:
                has_content = False

            request_count += 1
            if len(udfs) >= n or (
                max_requests is not None and request_count == max_requests
            ):
                break

        deserialized_udfs = {}
        for udf in udfs:
            if "udf_body" in udf and udf["udf_body"]:
                try:
                    deserialized_udf = RootAnyBaseUdf.model_validate_json(
                        udf["udf_body"]
                    ).root
                    udf_id = (
                        deserialized_udf.name
                        if by == "name"
                        else (
                            udf["id"]
                            if by == "id"
                            else (udf["slug"] if by == "slug" else None)
                        )
                    )

                    # Restore metadata fields if they were not already present
                    if not deserialized_udf._get_metadata_safe(METADATA_FUSED_ID):
                        deserialized_udf._set_metadata_safe(
                            METADATA_FUSED_ID, udf["id"]
                        )
                    if not deserialized_udf._get_metadata_safe(METADATA_FUSED_SLUG):
                        deserialized_udf._set_metadata_safe(
                            METADATA_FUSED_SLUG, udf["slug"]
                        )

                    filtered_public_udf = (
                        whose == "public"
                        and deserialized_udf._get_metadata_safe(
                            METADATA_FUSED_EXPLORER_TAB
                        )
                        == "community"
                    ) or (
                        whose == "community"
                        and deserialized_udf._get_metadata_safe(
                            METADATA_FUSED_EXPLORER_TAB
                        )
                        != "community"
                    )

                    if udf_id is not None and not filtered_public_udf:
                        deserialized_udfs[udf_id] = deserialized_udf
                except Exception as e:
                    warnings.warn(
                        f"UDF {udf['slug']} ({udf['id']}) could not be deserialized: {e}"
                    )
        return UdfRegistry(deserialized_udfs)

    def get_udf_access_tokens(
        self,
        n: Optional[int] = None,
        *,
        skip: int = 0,
        per_request: int = 25,
        max_requests: Optional[int] = 1,
        _whose: Literal["self", "all"] = "self",
    ) -> UdfAccessTokenList:
        request_count = 0
        has_content = True
        tokens = []

        assert per_request >= 0

        while has_content:
            url = f"{self.base_url}/udf-access-token/{_whose}"

            params = request_models.ListUdfAccessTokensRequest(
                skip=skip,
                limit=per_request,
            )
            skip += per_request

            r = requests.get(
                url=url,
                params=params.model_dump(),
                headers=self._generate_headers(),
                timeout=OPTIONS.request_timeout,
            )
            raise_for_status(r)
            tokens_this_request = r.json()
            if tokens_this_request:
                tokens.extend(tokens_this_request)
            else:
                has_content = False

            request_count += 1
            if n is not None and (
                len(tokens) >= n
                or (max_requests is not None and request_count == max_requests)
            ):
                break

        tokens_deserialized = UdfAccessTokenList()
        for token in tokens:
            tokens_deserialized.append(UdfAccessToken.model_validate(token))

        return tokens_deserialized

    def get_udf_access_token(
        self,
        token: Union[str, UdfAccessToken],
    ) -> UdfAccessToken:
        if isinstance(token, UdfAccessToken):
            token = token.token
        url = f"{self.base_url}/udf-access-token/by-token/{token}"

        r = requests.get(
            url=url,
            headers=self._generate_headers(),
            timeout=OPTIONS.request_timeout,
        )
        raise_for_status(r)
        return UdfAccessToken.model_validate_json(r.content)

    def delete_udf_access_token(
        self,
        token: Union[str, UdfAccessToken],
    ) -> UdfAccessToken:
        if isinstance(token, UdfAccessToken):
            token = token.token
        url = f"{self.base_url}/udf-access-token/by-token/{token}"

        r = requests.delete(
            url=url,
            headers=self._generate_headers(),
            timeout=OPTIONS.request_timeout,
        )
        raise_for_status(r)
        return UdfAccessToken.model_validate_json(r.content)

    def update_udf_access_token(
        self,
        token: Union[str, UdfAccessToken],
        *,
        client_id: Optional[str] = None,
        cache: Optional[bool] = None,
        metadata_json: Optional[Dict[str, Any]] = None,
        enabled: Optional[bool] = None,
    ) -> UdfAccessToken:
        if isinstance(token, UdfAccessToken):
            token = token.token
        url = f"{self.base_url}/udf-access-token/by-token/{token}"

        body = request_models.UpdateUdfAccessTokenRequest(
            client_id=client_id,
            cache=cache,
            metadata_json=metadata_json,
            enabled=enabled,
        ).model_dump()  # type: ignore

        r = requests.post(
            url=url,
            json=body,
            headers=self._generate_headers(),
            timeout=OPTIONS.request_timeout,
        )
        raise_for_status(r)
        return UdfAccessToken.model_validate_json(r.content)

    def create_udf_access_token(
        self,
        udf_email_or_name_or_id: Optional[str] = None,
        /,
        udf_name: Optional[str] = None,
        *,
        udf_email: Optional[str] = None,
        udf_id: Optional[str] = None,
        client_id: Union[str, Ellipsis, None] = ...,
        cache: bool = True,
        metadata_json: Optional[Dict[str, Any]] = None,
        enabled: bool = True,
    ) -> UdfAccessToken:
        """
        Create a token for running a UDF. The token allows anyone who has it to run
        the UDF, with the parameters they choose. The UDF will run under your environment.

        The token does not allow running any other UDF on your account.

        Args:
            udf_email_or_name_or_id: A UDF ID, email address (for use with udf_name), or UDF name.
            udf_name: The name of the UDF to create the

        Keyword Args:
            udf_email: The email of the user owning the UDF, or, if udf_name is None, the name of the UDF.
            udf_id: The backend ID of the UDF to create the token for.
            client_id: If specified, overrides which realtime environment to run the UDF under.
            cache: If True, UDF tiles will be cached.
            metadata_json: Additional metadata to serve as part of the tiles metadata.json.
            enable: If True, the token can be used.
        """
        if udf_id is not None:
            if (
                udf_name is not None
                or udf_email is not None
                or udf_email_or_name_or_id is not None
            ):
                warnings.warn(
                    "All other ways of specifying the UDF are ignored in favor of udf_id.",
                )
                udf_name = None
                udf_email = None
        elif udf_name is not None:
            if udf_email_or_name_or_id is not None:
                if udf_email is not None:
                    warnings.warn(
                        "All other ways of specifying the UDF are ignored in favor of the first argument and udf_name.",
                    )
                udf_email = udf_email_or_name_or_id
        elif udf_email_or_name_or_id is not None:
            if udf_name is not None:
                udf_email = udf_email_or_name_or_id
            else:
                # Need to figure out what exactly the first argument is and how it specifies a UDF
                is_valid_uuid = True
                try:
                    uuid.UUID(udf_email_or_name_or_id)
                except ValueError:
                    is_valid_uuid = False
                if is_valid_uuid:
                    udf_id = udf_email_or_name_or_id
                elif "/" in udf_email_or_name_or_id:
                    udf_email, udf_name = udf_email_or_name_or_id.split("/", maxsplit=1)
                else:
                    udf_name = udf_email_or_name_or_id
                    udf_email = self._whoami()["email"]
        else:
            raise ValueError("No UDF specified to create an access token for.")

        if client_id is Ellipsis:
            client_id = self._automatic_realtime_client_id()

        if client_id is Ellipsis:
            raise ValueError("Failed to detect realtime client ID")

        url = f"{self.base_url}/udf-access-token/new"

        metadata_json = metadata_json or {}

        body = request_models.CreateUdfAccessTokenRequest(
            udf_email=udf_email,
            udf_slug=udf_name,
            udf_id=udf_id,
            client_id=client_id,
            cache=cache,
            metadata_json=metadata_json,
            enabled=enabled,
        ).model_dump()  # type: ignore

        r = requests.post(
            url=url,
            json=body,
            headers=self._generate_headers(),
            timeout=OPTIONS.request_timeout,
        )
        raise_for_status(r)
        return UdfAccessToken.model_validate_json(r.content)

    def get_jobs(
        self,
        n: int = 5,
        *,
        skip: int = 0,
        per_request: int = 25,
        max_requests: Optional[int] = 1,
    ) -> Jobs:
        """Get the job history.

        Args:
            n: The number of jobs to fetch. Defaults to 5.

        Keyword Args:
            skip: Where in the job history to begin. Defaults to 0, which retrieves the most recent job.
            per_request: Number of jobs per request to fetch. Defaults to 25.
            max_requests: Maximum number of requests to make. May be None to fetch all jobs. Defaults to 1.

        Returns:
            The job history.
        """
        request_count = 0
        has_content = True
        jobs = []
        original_skip = skip

        assert per_request >= 0

        while has_content:
            url = f"{self.base_url}/job/self"

            params = request_models.ListJobsRequest(
                skip=skip,
                limit=per_request,
            )
            skip += per_request

            r = requests.get(
                url=url,
                params=params.model_dump(),
                headers=self._generate_headers(),
                timeout=OPTIONS.request_timeout,
            )
            raise_for_status(r)
            jobs_this_request = r.json()
            if jobs_this_request:
                jobs.extend(jobs_this_request)
            else:
                has_content = False

            request_count += 1
            if len(jobs) >= n or (
                max_requests is not None and request_count == max_requests
            ):
                break

        return Jobs(
            jobs=jobs[:n],
            n=n,
            skip=original_skip,
            per_request=per_request,
            max_requests=max_requests,
        )

    def get_job_config(self, job: CoerceableToJobId) -> JobConfig:
        job_id = _object_to_job_id(job)
        url = f"{self.base_url}/job/by-id/{job_id}/config"

        r = requests.get(
            url=url,
            headers=self._generate_headers(),
            timeout=OPTIONS.request_timeout,
            allow_redirects=False,
        )
        raise_for_status(r)

        redirect_location = r.headers["location"]

        r2 = requests.get(
            url=redirect_location,
            timeout=OPTIONS.request_timeout,
        )
        raise_for_status(r2)

        return JobConfig.model_validate_json(r2.content)

    def get_status(self, job: CoerceableToJobId) -> RunResponse:
        """Fetch the status of a running job

        Args:
            job: the identifier of a job or a `RunResponse` object.

        Returns:
            The status of the given job.
        """
        job_id = _object_to_job_id(job)
        url = f"{self.base_url}/run/by-id/{job_id}"

        r = requests.get(
            url=url,
            headers=self._generate_headers(),
            timeout=OPTIONS.request_timeout,
        )
        raise_for_status(r)
        return RunResponse.model_validate_json(r.content)

    def get_logs(
        self,
        job: CoerceableToJobId,
        since_ms: Optional[int] = None,
    ) -> List[Any]:
        """Fetch logs for a job

        Args:
            job: the identifier of a job or a `RunResponse` object.
            since_ms: Timestamp, in milliseconds since epoch, to get logs for. Defaults to None for all logs.

        Returns:
            Log messages for the given job.
        """
        job_id = _object_to_job_id(job)
        url = f"{self.base_url}/logs/{job_id}"
        params = {}
        if since_ms is not None:
            params["since_ms"] = since_ms
        r = requests.get(
            url=url,
            params=params,
            headers=self._generate_headers(),
            timeout=OPTIONS.request_timeout,
        )
        raise_for_status(r)
        return r.json()

    def tail_logs(
        self,
        job: CoerceableToJobId,
        refresh_seconds: float = 1,
        sample_logs: bool = True,
        timeout: Optional[float] = None,
        get_logs_retries: int = 1,
    ):
        """Continuously print logs for a job

        Args:
            job: the identifier of a job or a `RunResponse` object.
            refresh_seconds: how frequently, in seconds, to check for new logs. Defaults to 1.
            sample_logs: if true, print out only a sample of logs. Defaults to True.
            timeout: if not None, how long to continue tailing logs for. Defaults to None for indefinite.
            get_logs_retries: Number of additional retries for log requests. Defaults to 1.
        """
        # TODO: Move this to the RunResponse object
        start_time = time.time()
        job = self.get_status(job)
        print(f"Logs for: {job.job_id}")

        def _tail_get_logs(
            job: RunResponse, since_ms: Optional[int] = None
        ) -> List[Any]:
            for _ in range(get_logs_retries + 1):
                try:
                    return self.get_logs(job, since_ms=since_ms)
                except requests.exceptions.RequestException:
                    # TODO: Don't use bare except
                    print("Server did not respond with logs")
                    # TODO: Backoff strategy
                    time.sleep(refresh_seconds)
            raise ValueError("Server did not respond with logs")

        r = _tail_get_logs(job)
        if job.status not in ["running", "pending"]:
            print(f"Job is not running ({job.status})")
            return

        if len(r) == 0:
            print("Configuring packages and waiting for logs...")
            while len(r) == 0:
                time.sleep(refresh_seconds)
                r = _tail_get_logs(job)
                if timeout is not None and time.time() - start_time > timeout:
                    raise TimeoutError("Timed out waiting for logs")

        last_message: Optional[str] = None
        last_since_ms: Optional[int] = None
        while True:
            time.sleep(refresh_seconds)
            r = _tail_get_logs(job, since_ms=last_since_ms)
            # If any results -- there may be none because we are filtering them with since_ms
            if len(r):
                current_message: str = r[-1]["message"]
                if last_message != current_message:
                    # If the most recent log line has changed, print it out
                    last_message = current_message
                    last_since_ms = r[-1]["timestamp"]

                    if sample_logs:
                        print(current_message.rstrip())
                    else:
                        for message in r:
                            print(message["message"].rstrip())

            if "ERROR" in current_message or self.get_status(job).status != "running":
                # Try to detect exit scenarios: an error has occured and the job will stop,
                # or the job is no longer in a running state.
                return

            if timeout is not None and time.time() - start_time > timeout:
                raise TimeoutError("Timed out")

    def wait_for_job(
        self,
        job: CoerceableToJobId,
        poll_interval_seconds: float = 5,
        timeout: Optional[float] = None,
    ) -> RunResponse:
        """Block the Python kernel until the given job has finished

        Args:
            job: the identifier of a job or a `RunResponse` object.
            poll_interval_seconds: How often (in seconds) to poll for status updates. Defaults to 5.
            timeout: The length of time in seconds to wait for the job. Defaults to None.

        Raises:
            TimeoutError: if waiting for the job timed out.

        Returns:
            The status of the given job.
        """
        # TODO: Move this to the RunResponse object
        start_time = time.time()
        status = self.get_status(job)
        while not status.terminal_status:
            time.sleep(poll_interval_seconds)
            status = self.get_status(job)
            if timeout is not None and time.time() - start_time > timeout:
                raise TimeoutError("Timed out waiting for job")
        return status

    def cancel_job(self, job: CoerceableToJobId) -> RunResponse:
        """Cancel an existing job

        Args:
            job: the identifier of a job or a `RunResponse` object.

        Returns:
            A new job object.
        """
        job_id = _object_to_job_id(job)
        url = f"{self.base_url}/run/by-id/{job_id}/cancel"

        self._check_is_prod()
        r = requests.post(
            url=url,
            headers=self._generate_headers(),
            timeout=OPTIONS.request_timeout,
        )
        raise_for_status(r)
        return RunResponse.model_validate_json(r.content)

    def _whoami(self) -> Any:
        """
        Returns information on the currently logged in user
        """
        url = f"{self.base_url}/user/self"

        r = requests.get(
            url=url,
            headers=self._generate_headers(),
            timeout=OPTIONS.request_timeout,
        )
        raise_for_status(r)
        return r.json()

    def _list_realtime_instances(self, *, whose: str = "self") -> List[Any]:
        """
        Returns information about available realtime instances
        """
        url = f"{self.base_url}/realtime-instance"
        if whose == "self":
            url += "/available"
        else:
            assert whose == "public", "whose must be 'public' or 'self'"

        r = requests.get(
            url=url,
            headers=self._generate_headers(),
            timeout=OPTIONS.request_timeout,
        )
        raise_for_status(r)
        return r.json()

    def _automatic_realtime_client_id(self) -> Optional[str]:
        client_id = OPTIONS.realtime_client_id
        if client_id is None:
            instances = self._list_realtime_instances()
            if len(instances):
                client_id = instances[0]["client_id"]
            if OPTIONS.save_user_settings and client_id:
                OPTIONS.realtime_client_id = client_id

        return client_id

    def open_table(
        self,
        path: str,
        *,
        fetch_samples: Optional[bool] = None,
    ) -> Table:
        """Open a Table object given a path to the root of the table

        Args:
            path: The path to the root of the table on remote storage

        Keyword Args:
            fetch_samples: If True, fetch sample of each table when opening the dataset.

        Example:

            ```python
            table = fused.open_table(path="s3://my_bucket/path/to/dataset/table/")
            ```

        Returns:
            A Table object
        """
        request = request_models.OpenTableRequest(
            path=path,  # type: ignore
        )
        url = f"{self.base_url}/table/open"
        r = requests.get(
            url=url,
            headers=self._generate_headers(),
            params=request.model_dump(),
            timeout=OPTIONS.request_timeout,
        )
        raise_for_status(r)
        table = Table.model_validate(r.json())
        post_open_table(table=table, fetch_samples=fetch_samples)

        return table

    def open_folder(
        self,
        path: str,
        *,
        fetch_minimal_table_metadata: Optional[bool] = None,
        fetch_table_metadata: Optional[bool] = None,
        fetch_samples: Optional[bool] = None,
        table_mode: bool = True,
        max_depth: Optional[int] = None,
    ) -> Folder:
        """Open all Table objects under the path.

        Args:
            path: The path to the root of the folder on remote storage

        Keyword Args:
            fetch_table_metadata: If True, fetch metadata on each table when getting dataset metadata.
            fetch_samples: If True, fetch sample of each table when opening the dataset.
            max_depth: Maximum depth of Tables to open. Beyond this tables will be opened virtually.

        Example:

            datasets = fused.open_folder(
                path="s3://my_bucket/path/to/folder/"
            )

        Returns:
            A list of Dataset objects
        """
        should_fetch_table_metadata = False
        if fetch_table_metadata is not None:
            should_fetch_table_metadata = fetch_table_metadata
        elif OPTIONS.open.fetch_table_metadata is not None:
            should_fetch_table_metadata = OPTIONS.open.fetch_table_metadata

        should_fetch_minimal_table_metadata = True
        if fetch_minimal_table_metadata is not None:
            should_fetch_minimal_table_metadata = fetch_minimal_table_metadata
        elif OPTIONS.open.fetch_minimal_table_metadata is not None:
            should_fetch_minimal_table_metadata = (
                OPTIONS.open.fetch_minimal_table_metadata
            )

        request = request_models.OpenDatasetFolderRequest(
            path=path,  # type: ignore
            fetch_minimal_table_metadata=should_fetch_minimal_table_metadata,
            fetch_table_metadata=should_fetch_table_metadata,
            max_depth=max_depth,
        )
        mode_str = "table" if table_mode else "dataset"
        url = f"{self.base_url}/{mode_str}/open_folder"
        r = requests.get(
            url=url,
            headers=self._generate_headers(),
            params=request.model_dump(),
            timeout=OPTIONS.request_timeout,
        )
        raise_for_status(r)
        folder = Folder.model_validate(r.json())
        for table in folder.tables:
            post_open_table(
                table=table,
                fetch_samples=fetch_samples,
            )

        return folder

    @overload
    def list(self, path: str, *, details: Literal[True]) -> List[ListDetails]:
        ...

    @overload
    def list(self, path: str, *, details: Literal[False] = False) -> List[str]:
        ...

    def list(self, path: str, *, details: bool = False):
        list_request_url = f"{self.base_url}/files/list{'-details' if details else ''}"
        params = request_models.ListPathRequest(path=path)
        r = requests.get(
            url=list_request_url,
            params=params.model_dump(),
            headers=self._generate_headers(),
            timeout=OPTIONS.request_timeout,
        )
        raise_for_status(r)
        result = r.json()
        if details:
            result = [ListDetails.model_validate(detail) for detail in result]
        return result

    def delete(
        self,
        path: str,
        max_deletion_depth: Union[int, Literal["unlimited"]] = 2,
    ) -> bool:
        delete_request_url = f"{self.base_url}/files/delete"

        params = request_models.DeletePathRequest(
            path=path,
            max_deletion_depth=max_deletion_depth,
        )

        r = requests.delete(
            url=delete_request_url,
            params=params.model_dump(),
            json="{}",
            headers=self._generate_headers(),
            timeout=OPTIONS.request_timeout,
        )
        raise_for_status(r)
        return r.json()

    def _resolve(
        self,
        path: str,
    ) -> List[str]:
        resolve_request_url = f"{self.base_url}/files/resolve"

        params = request_models.ResolvePathRequest(
            path=path,
        )

        r = requests.post(
            url=resolve_request_url,
            params=params.model_dump(),
            headers=self._generate_headers(),
            timeout=OPTIONS.request_timeout,
        )
        raise_for_status(r)
        return r.json()

    def get(self, path: str) -> bytes:
        get_request_url = f"{self.base_url}/files/get"
        params = request_models.GetPathRequest(path=path)
        r = requests.get(
            url=get_request_url,
            params=params.model_dump(),
            headers=self._generate_headers(),
            timeout=OPTIONS.request_timeout,
            allow_redirects=True,
        )
        raise_for_status(r)
        return r.content

    def download(self, path: str, local_path: Union[str, Path]) -> None:
        get_request_url = f"{self.base_url}/files/get"
        params = request_models.GetPathRequest(path=path)
        r = requests.get(
            url=get_request_url,
            params=params.model_dump(),
            headers=self._generate_headers(),
            timeout=OPTIONS.request_timeout,
            allow_redirects=True,
            stream=True,
        )
        raise_for_status(r)
        with open(local_path, "wb") as f:
            shutil.copyfileobj(r.raw, f)

    def sign_url(self, path: str) -> str:
        sign_request_url = f"{self.base_url}/files/sign"
        params = request_models.SignPathRequest(path=path)
        r = requests.get(
            url=sign_request_url,
            params=params.model_dump(),
            headers=self._generate_headers(),
            timeout=OPTIONS.request_timeout,
        )
        raise_for_status(r)
        return r.json()

    def sign_url_prefix(self, path: str) -> Dict[str, str]:
        sign_prefix_request_url = f"{self.base_url}/files/sign_prefix"
        params = request_models.SignPathRequest(path=path)
        r = requests.get(
            url=sign_prefix_request_url,
            params=params.model_dump(),
            headers=self._generate_headers(),
            timeout=OPTIONS.request_timeout,
        )
        raise_for_status(r)
        return r.json()

    def _get_table_names(self, path: str) -> List[str]:
        list_request_url = f"{self.base_url}/files/list"
        params = request_models.ListPathRequest(path=path)
        r = requests.get(
            url=list_request_url,
            params=params.model_dump(),
            headers=self._generate_headers(),
            timeout=OPTIONS.request_timeout,
        )
        raise_for_status(r)

        tables: List[str] = []
        for s3_path in r.json():
            table_name = s3_path.rstrip("/").split("/")[-1]
            tables.append(table_name)

        return tables

    def upload(self, path: str, data: Union[bytes, BinaryIO]) -> None:
        """Upload a binary blob to a cloud location"""
        upload_url = f"{self.base_url}/files/upload"
        params = request_models.UploadRequest(path=path)
        r = requests.put(
            url=upload_url,
            params=params.model_dump(),
            headers=self._generate_headers(
                {"Content-Type": "application/octet-stream"}
            ),
            data=data,
            timeout=OPTIONS.request_timeout,
        )
        raise_for_status(r)

    def _upload_tmp(self, extension: str, data: Union[bytes, BinaryIO]) -> str:
        """Upload a binary blob to a temporary cloud location, and return the new URL"""
        upload_temp_url = f"{self.base_url}/files/upload-temp"
        params = request_models.UploadTempRequest(extension=extension)
        r = requests.post(
            url=upload_temp_url,
            params=params.model_dump(),
            headers=self._generate_headers(
                {"Content-Type": "application/octet-stream"}
            ),
            data=data,
            timeout=OPTIONS.request_timeout,
        )
        raise_for_status(r)
        return r.json()

    def _replace_df_input(
        self, input: Union[str, List[str], Path, gpd.GeoDataFrame]
    ) -> Union[str, List[str]]:
        """If the input is a DataFrame, upload it and return a URL to it. Otherwise return input unchanged."""
        if isinstance(input, gpd.GeoDataFrame):
            with TemporaryFile() as tmp:
                input.to_parquet(tmp)
                tmp.seek(0, SEEK_SET)
                input = self._upload_tmp(extension="parquet", data=tmp)
        elif isinstance(input, Path):
            with open(input, "rb") as f:
                extension = input.name.rsplit(".", 1)[-1]
                input = self._upload_tmp(extension=extension, data=f)
        return input

    def _health(self) -> bool:
        """Check the health of the API backend"""
        r = requests.get(f"{self.base_url}/health", timeout=OPTIONS.request_timeout)
        raise_for_status(r)
        return True

    def auth_token(self) -> str:
        """
        Returns the current user's Fused environment (team) auth token
        """
        url = f"{self.base_url}/execution-env/token"

        r = requests.get(
            url=url,
            headers=self._generate_headers(),
            timeout=OPTIONS.request_timeout,
        )
        raise_for_status(r)
        return r.json()

    def _generate_headers(
        self,
        headers: Optional[Dict[str, str]] = None,
        *,
        credentials_needed: bool = True,
    ) -> Dict[str, str]:
        if headers is None:
            headers = {}

        common_headers = {
            "Fused-Py-Version": fused_batch.__version__,
            **headers,
        }

        if AUTHORIZATION.is_configured() or credentials_needed:
            common_headers[
                "Authorization"
            ] = f"{AUTHORIZATION.credentials.auth_scheme} {AUTHORIZATION.credentials.access_token}"

        return common_headers

    @lru_cache
    def dependency_whitelist(self) -> str:
        sign_request_url = f"{self.base_url}/internal/dependency-whitelist"
        r = requests.get(
            url=sign_request_url,
            headers=self._generate_headers(),
            timeout=OPTIONS.request_timeout,
        )
        raise_for_status(r)
        return r.json()


set_api_class(FusedAPI)



from __future__ import annotations

import warnings
from concurrent.futures import ThreadPoolExecutor
from dataclasses import dataclass
from datetime import datetime
from itertools import zip_longest
from typing import TYPE_CHECKING, Dict, Generator, Iterable, List, Optional, Sequence

import geopandas as gpd

from fused_batch._options import options as OPTIONS

if TYPE_CHECKING:
    from pystac import Item


__all__ = ("stac_from_tiff_list", "stac_from_stac_list")


@dataclass
class IterInput:
    tiff_url: str
    id: Optional[str]
    input_datetime: Optional[datetime]


def stac_from_tiff_list(
    tiff_list: Sequence[str],
    *,
    input_datetime: Optional[Iterable[datetime]] = None,
    id: Optional[Iterable[str]] = None,
    collection: Optional[str] = None,
    asset_name: str = "asset",
    naive: bool = False,
    max_workers: Optional[int] = None,
) -> gpd.GeoDataFrame:
    """Construct a STAC GeoDataFrame from a list of GeoTIFF urls.

    Args:
        tiff_list: input paths.

    Keyword Args:
        input_datetime: datetime associated with the item. Defaults to None.
        id: id to assign to the item (default to the source basename). Defaults to None.
        collection: name of collection the item belongs to. Defaults to None.
        asset_name: asset name in the Assets object. Defaults to "asset".
        naive: When True, this function will only read geometry information for the _first_ item in the list and copy geometry information to the others. Defaults to False.

    Returns:
        a GeoDataFrame with containing the description of STAC items
    """
    from rio_stac import create_stac_item

    iterator = _iter_input(tiff_list, input_datetime=input_datetime, id=id)
    first_item = next(iterator)

    first_stac = create_stac_item(
        first_item.tiff_url,
        id=first_item.id,
        input_datetime=first_item.input_datetime,
        collection=collection,
        asset_name=asset_name,
        with_proj=True,
        with_eo=True,
    )

    if naive:
        return _infer_stac_naive(
            first_item=first_stac, iterator=iterator, asset_name=asset_name
        )

    stac_list: List[Dict] = [first_stac.to_dict()]

    def _mapper(_input_item: IterInput) -> Dict:
        return create_stac_item(
            _input_item.tiff_url,
            id=_input_item.id,
            input_datetime=_input_item.input_datetime,
            collection=collection,
            asset_name=asset_name,
            with_proj=True,
            with_eo=True,
        ).to_dict()

    max_workers = max_workers if max_workers is not None else OPTIONS.max_workers
    with ThreadPoolExecutor(max_workers=max_workers) as pool:
        stac_list.extend(pool.map(_mapper, iterator))

    return stac_from_stac_list(stac_list)


def _infer_stac_naive(
    first_item: Item, iterator: Generator[IterInput, None, None], asset_name: str
) -> gpd.GeoDataFrame:
    warnings.warn(
        "Using naive STAC generation, copying geometries from first GeoTIFF in the list."
    )

    stac_list: List[Dict] = [first_item.to_dict()]

    for _input in iterator:
        new_item = first_item.clone()
        new_item.id = _input.id or _input.tiff_url.rsplit("/")[-1]
        new_item.assets[asset_name].href = _input.tiff_url
        if _input.input_datetime:
            new_item.datetime = _input.input_datetime

        stac_list.append(new_item.to_dict())

    return stac_from_stac_list(stac_list)


def _iter_input(
    tiff_list: List[str],
    *,
    input_datetime: Optional[Iterable[datetime]] = None,
    id: Optional[Iterable[str]] = None,
) -> Generator[IterInput, None, None]:
    # Use itertools.zip_longest
    for tiff_url, _input_datetime, _id in zip_longest(
        tiff_list, input_datetime or [], id or []
    ):
        yield IterInput(tiff_url=tiff_url, id=_id, input_datetime=_input_datetime)


def stac_from_stac_list(stac_list: Sequence[Dict]) -> gpd.GeoDataFrame:
    """Construct a STAC GeoDataFrame from a list of STAC Items.

    Args:
        stac_list: input STAC Items.

    Returns:
        a GeoDataFrame with containing the description of STAC items
    """
    import stac_geoparquet

    return stac_geoparquet.to_geodataframe(stac_list)



def data_path() -> str:
    # TODO: Allow overriding in OPTIONS
    return "/tmp/fused"


def filesystem(protocol: str, **storage_options):
    from urllib.parse import urlparse

    import fsspec
    from fsspec.implementations.dirfs import DirFileSystem

    from fused.api._public_api import get_api

    if protocol == "fd":
        # fused team directory
        api = get_api()
        if hasattr(api, "_resolve"):
            root = api._resolve("fd://")
            root_parsed = urlparse(root)
            return DirFileSystem(
                path=root, fs=fsspec.filesystem(root_parsed.scheme, **storage_options)
            )
        else:
            raise ValueError("Could not determine root of Fused team directory")
    return fsspec.filesystem(protocol, **storage_options)


def _download_requests(url: str) -> bytes:
    import requests

    # this function is shared
    response = requests.get(url, headers={"User-Agent": ""})
    response.raise_for_status()
    return response.content


def _download_signed(url: str) -> bytes:
    from fused.api._public_api import get_api

    api = get_api()
    return _download_requests(api.sign_url(url))


def _download_s3(url: str) -> bytes:
    try:
        return _download_signed(url)
    except:  # noqa E722
        s3 = filesystem("s3")
        with s3.open(url, "rb") as f:
            return f.read()


def _download_gcs(url: str) -> bytes:
    try:
        return _download_signed(url)
    except:  # noqa E722
        gcs = filesystem("gs")
        with gcs.open(url, "rb") as f:
            return f.read()



# ruff: noqa: F401

from . import _raster as raster
from . import _templates as templates
from ._project import Project, open_project
from ._public_api import get_jobs, job, join, load_job, map, open_table
from ._public_api import plot as _plot
from ._public_api import sign_url, sign_url_prefix, union_tables, upload, zip_tables
from ._udf import load_udf



import concurrent.futures
import time
from typing import Optional

from requests.exceptions import ReadTimeout

from fused._optional_deps import HAS_PANDAS, PD_DATAFRAME
from fused._run import run as fused_run
from fused.models.udf.udf import GeoPandasUdfV2
from fused.types import UdfRuntimeError


class Future:
    # or `Job`, `JobResult`, `Task`, ...

    def __init__(self, future):
        self._future: concurrent.futures.Future = future

    def done(self):
        return self._future.done()

    def result(self):
        response = self._future.result()
        if isinstance(response, Exception):
            raise response
        if response.error_message is not None:
            raise self.exception()
        return response.data

    def exception(self):
        response = self._future.result()
        if isinstance(response, Exception):
            return response
        if response.error_message is not None:
            return UdfRuntimeError(
                response.error_message,
                child_exception_class=response.exception_class,
            )
        return self._future.exception()

    def get_logs(self):
        response = self._future.result()
        if isinstance(response, ReadTimeout):
            return str(response)
        out = ""
        if response.stdout:
            out += "stdout\n------\n" + response.stdout
        if response.stderr:
            out += "\nstderr\n------\n" + response.stderr
        return out

    def status(self):
        if self._future.running():
            return "running"
        elif self._future.done():
            exc = self.exception()
            if exc:
                if isinstance(
                    exc, ReadTimeout
                ) or "504 Server Error: Gateway Timeout" in str(exc):
                    return "timeout"
                return "error"
            return "success"  # or "finished"?
        else:
            return "pending"

    def __repr__(self):
        return f"<fused.Future [status: {('done - ' if self.done() else '') + self.status()}]>"


class JobPool:
    # or `Futures`, `PoolRunner`, etc

    """
    Pool of UDF runs. Don't use this class directly, use `fused.submit` instead.
    """

    # TODO add cancel support
    # TODO add retry support

    def __init__(
        self,
        udf,
        arg_list,
        kwargs=None,
        engine="remote",
        max_workers=None,
        max_retry=2,
    ):
        self.udf = udf
        self.arg_list = arg_list
        self.n_jobs = len(self.arg_list)
        self._kwargs = kwargs or {}
        self.engine = engine
        max_workers = min(max_workers or 32, 1024)
        self._pool = concurrent.futures.ThreadPoolExecutor(max_workers=max_workers)
        self._max_retry = max_retry

    def _start_jobs(self):
        def _run(args):
            # currently we have to add a small delay between starting the UDF runs
            # to avoid overloading the server
            time.sleep(0.01)
            try:
                return fused_run(
                    self.udf,
                    engine=self.engine,
                    _return_response=True,
                    max_retry=self._max_retry,
                    **args,
                    **self._kwargs,
                )
            except Exception as exc:
                # ReadTime or HTTPError can happen on time-out or other server error
                return exc

        self._futures = [
            Future(self._pool.submit(_run, args)) for args in self.arg_list
        ]

    def done(self):
        return all(fut.done() for fut in self._futures)

    def _get_status(self):
        return [f.done() for f in self._futures]

    def _get_progress(self):
        percentage = round(sum(self._get_status()) / self.n_jobs * 100)
        return percentage, f"{sum(self._get_status())}/{self.n_jobs}"

    def wait(self):
        from tqdm.auto import tqdm

        t = tqdm(total=self.n_jobs, bar_format="{l_bar}{bar} | {n_fmt}/{total_fmt}")
        while not self.done():
            n_done = sum(self._get_status())
            t.update(n_done - t.n)
        t.update(self.n_jobs - t.n)
        t.close()

    def results(self, return_exceptions=False):
        results = []
        for fut in self._futures:
            try:
                results.append(fut.result())
            except Exception:
                if return_exceptions:
                    results.append(fut.exception())
                else:
                    raise
        return results

    def __getitem__(self, idx: int) -> Future:
        return self._futures[idx]

    def __len__(self):
        return self.n_jobs

    def __repr__(self):
        # TODO we could provide a more informative repr in notebooks (e.g. showing
        # a table of the individual jobs and their status?)
        return f"<JobPool with {self.n_jobs} jobs [status: {self._get_progress()[1]} done]>"

    def get_status_df(self):
        if not HAS_PANDAS:
            raise ImportError("pandas is required to use this method")

        import pandas as pd

        results_df = pd.DataFrame(self.arg_list)
        results_df["status"] = [f.status() for f in self._futures]
        results_df["result"] = "<pending>"
        for i, fut in enumerate(self._futures):
            if results_df["status"].iloc[i] not in ("running", "pending"):
                try:
                    res = fut.result()
                except Exception:
                    res = fut.exception()
                results_df.at[i, "result"] = res

        return results_df

    def get_results_df(self, ignore_exceptions=False):
        if not HAS_PANDAS:
            raise ImportError("pandas is required to use this method")

        import pandas as pd

        results = self.results(return_exceptions=True)
        results_df = pd.DataFrame(self.arg_list)
        results_df["status"] = [f.status() for f in self._futures]
        results_df["result"] = results
        if ignore_exceptions:
            mask = [not isinstance(r, Exception) for r in results]
            results_df = results_df[mask]
        return results_df

    def collect_df(self, ignore_exceptions=True):
        if not HAS_PANDAS:
            raise ImportError("pandas is required to use this method")

        import pandas as pd

        results = self.results(return_exceptions=ignore_exceptions)
        results_df = pd.concat(
            dict(
                res
                for res in zip(range(self.n_jobs), results)
                if not isinstance(res[1], Exception)
            )
        )

        args_df = pd.DataFrame(self.arg_list)
        mask = [not isinstance(r, Exception) for r in results]
        args_df = args_df[mask]

        # combine concatenated results with arguments as prepended index levels
        assert len(results_df.index.levels[0]) == len(args_df)
        args_index = pd.MultiIndex.from_frame(args_df)
        args_codes = [c.take(results_df.index.codes[0]) for c in args_index.codes]
        new_idx = pd.MultiIndex(
            levels=list(args_index.levels) + results_df.index.levels[1:],
            codes=args_codes + results_df.index.codes[1:],
            names=list(args_df.columns) + results_df.index.names[1:],
        )
        results_df.index = new_idx
        return results_df


def _validate_arg_list(arg_list, udf):
    if HAS_PANDAS and isinstance(arg_list, PD_DATAFRAME):
        return arg_list.to_dict(orient="records")

    if not len(arg_list):
        raise ValueError("arg_list must be a non-empty list")
    if not isinstance(arg_list[0], dict):
        if not isinstance(udf, GeoPandasUdfV2):
            raise ValueError(
                "arg_list must be a list of dictionaries. A simple list to pass "
                "as first positional argument is only supported for UDF objects."
            )
        name = udf._parameter_list[0]
        arg_list = [{name: arg} for arg in arg_list]

    return arg_list


def submit(
    udf,
    arg_list,
    /,
    *,
    engine: str = "remote",
    max_workers: Optional[int] = None,
    max_retry: int = 2,
    debug_mode: bool = False,
    wait_on_results: bool = False,
    **kwargs,
):
    """
    Executes a user-defined function (UDF) multiple times for a list of input
    parameters, and return immediately a "lazy" JobPool object allowing
    to inspect the jobs and wait on the results.

    See `fused.run` for more details on the UDF execution.

    Args:
        udf: the UDF to execute.
            See `fused.run` for more details on how tos specify the UDF.
        arg_list: a list of input parameters for the UDF. Can be specified as:
            - a list of values for parametrizing over a single parameter, i.e.
              the first parameter of the UDF
            - a list of dictionaries for parametrizing over multiple parameters
            - A DataFrame for parametrizing over multiple parameters where each
              row is a set of parameters

    Returns:
        JobPool

    Examples:

        Run a UDF multiple times for the values 0 to 9 passed to as the first
        positional argument of the UDF:
        ```py
        pool = fused.submit("username@fused.io/my_udf_name", range(10))
        ```

        Being explicit about the parameter name:
        ```py
        pool = fused.submit(udf, [dict(n=i) for i in range(10)])
        ```

    """
    arg_list = _validate_arg_list(arg_list, udf)

    if debug_mode:
        return fused_run(udf, engine=engine, **arg_list[0], **kwargs)

    job_pool = JobPool(
        udf,
        arg_list,
        kwargs,
        engine=engine,
        max_workers=max_workers,
        max_retry=max_retry,
    )
    job_pool._start_jobs()

    if wait_on_results:
        job_pool.wait()
        return job_pool.get_results_df()
    else:
        return job_pool



from typing import Optional, Protocol, Union

import geopandas as gpd
import pandas as pd

from fused_batch.models.schema import Schema
from fused_batch.models.udf.output import Output as OutputModel


class Output(Protocol):
    @property
    def data(self) -> Union[pd.DataFrame, gpd.GeoDataFrame, None]:
        ...

    @data.setter
    def data(self, value: Union[pd.DataFrame, gpd.GeoDataFrame, None]):
        ...

    @property
    def table_schema(self) -> Optional[Schema]:
        ...

    @table_schema.setter
    def table_schema(self, value: Optional[Schema]):
        ...


output: Output = OutputModel()



import warnings
from typing import Any, Callable, Optional

from fused_batch.warnings import FusedWarning


def noraise(
    fn: Optional[Callable] = None,
    *,
    incompat_version_message: Optional[str] = None,
    message: Optional[str] = None,
    default: Optional[Any] = None,
):
    rendered_message = "Failed, and no additional information is available."
    if message:
        rendered_message = message
    elif incompat_version_message:
        rendered_message = f"{incompat_version_message}. You may have an incompatible version of fused-py -- try upgrading to latest."

    def _internal_noraise_wrapper(func):
        def _internal_noraise(*args, **kwargs):
            try:
                return func(*args, **kwargs)
            except:  # noqa: E722
                warnings.warn(rendered_message, FusedWarning)
                return default

        return _internal_noraise

    if fn is not None:
        return _internal_noraise_wrapper(fn)
    else:
        return _internal_noraise_wrapper



from __future__ import annotations

import secrets
from textwrap import dedent
from typing import Any, Dict
from urllib.parse import urlencode

import requests

from fused_batch._auth import (
    AUDIENCE,
    AUTHORIZATION,
    AUTHORIZE_URL,
    CLIENT_ID,
    CLIENT_SECRET,
    OAUTH_TOKEN_URL,
    SCOPES,
    Credentials,
    delete_token_from_disk,
    get_code_challenge,
)
from fused_batch._auth import logout as auth_logout
from fused_batch._global_api import get_api, reset_api
from fused_batch._options import options as OPTIONS

HOSTED_REDIRECT_URI = "https://www.fused.io/notebook_login_redirect"


class NotebookCredentials:
    """
    To use this credentials helper, run the following and it will guide you to create a FusedAPI object.

    ```py
    credentials = NotebookCredentials()
    ```
    """

    _code_verifier: str
    _api_kwargs: Dict[str, Any]
    _show_widget: bool
    url: str

    def __init__(
        self, *, api_kwargs: Dict[str, Any] | None = None, show_widget: bool = True
    ):
        self._code_verifier = secrets.token_urlsafe(48)
        self._api_kwargs = api_kwargs or {}
        self._show_widget = show_widget
        code_challenge = get_code_challenge(self._code_verifier)

        params = {
            "audience": AUDIENCE,
            "scope": " ".join(SCOPES),
            "response_type": "code",
            "client_id": CLIENT_ID,
            "redirect_uri": HOSTED_REDIRECT_URI,
            "code_challenge_method": "S256",
            "code_challenge": code_challenge,
        }
        self.url = f"{AUTHORIZE_URL}?{urlencode(params)}"
        if self._show_widget:
            from IPython.core.display import Markdown
            from IPython.display import display

            login_url_message = f"""
                Please open this link to authenticate with Fused:

                * [{self.url}]({self.url})

                Once you have logged in, that page will give you a code which you can paste into this notebook to finish logging in.
                """
            display(Markdown(dedent(login_url_message)))

    def finalize(self, code):
        token_data = {
            "client_id": CLIENT_ID,
            "grant_type": "authorization_code",
            "audience": AUDIENCE,
            "client_secret": CLIENT_SECRET,
            "code": code,
            "redirect_uri": HOSTED_REDIRECT_URI,
            "code_verifier": self._code_verifier,
        }
        token_response = requests.post(
            OAUTH_TOKEN_URL, json=token_data, timeout=OPTIONS.request_timeout
        )
        token_response.raise_for_status()
        token_dict = token_response.json()

        credentials = Credentials.from_token_response(token_dict)
        AUTHORIZATION.set_credentials(credentials)

        if self._show_widget:
            from IPython.core.display import Markdown
            from IPython.display import display

            logged_in_message = "You are now logged in to Fused. :rocket:"
            display(Markdown(logged_in_message))

        return get_api()


def logout():
    """Log out the current user.

    This deletes the credentials saved to disk and resets the global Fused API.
    """
    delete_token_from_disk()
    auth_logout()
    reset_api()


def access_token() -> str:
    """Get an access token for the Fused service."""
    return AUTHORIZATION.credentials.access_token


def auth_scheme() -> str:
    return AUTHORIZATION.credentials.auth_scheme
